"""
Computation of alpha shape algorithm in 2-D based on original implementation
by Tim Kittel (@timkittel) available at:

    https://github.com/timkittel/alpha-shapes

Author(s):
    Dani Arribas-Bel daniel.arribas.bel@gmail.com
    Levi John Wolf levi.john.wolf@gmail.com
"""

import numpy as np
import scipy.spatial as spat
from packaging.version import Version
from scipy import sparse

from ..common import HAS_JIT, jit, requires

if not HAS_JIT:
    from warnings import warn

    NUMBA_WARN = (
        "Numba not imported, so alpha shape construction may be slower than expected."
    )

try:
    import shapely

    assert Version(shapely.__version__) >= Version("2")

    HAS_SHAPELY = True
except (ModuleNotFoundError, AssertionError):
    HAS_SHAPELY = False


EPS = np.finfo(float).eps

__all__ = ["alpha_shape", "alpha_shape_auto"]


@jit(nopython=True)
def nb_dist(x, y):
    """numba implementation of distance between points `x` and `y`

    Parameters
    ----------
    x : ndarray
        Coordinates of point `x`
    y : ndarray
        Coordinates of point `y`

    Returns
    -------
    dist : float
        Distance between `x` and `y`

    Examples
    --------
    >>> x = np.array([0, 0])
    >>> y = np.array([1, 1])
    >>> dist = nb_dist(x, y)
    >>> dist
    1.4142135623730951
    """
    sum_ = 0
    for x_i, y_i in zip(x, y):  # noqa B905
        sum_ += (x_i - y_i) ** 2
    dist = np.sqrt(sum_)
    return dist


@jit(nopython=True)
def r_circumcircle_triangle_single(a, b, c):
    """Computation of the circumcircle of a single triangle

    Parameters
    ----------
    a : ndarray
        (2,) Array with coordinates of vertex `a` of the triangle
    b : ndarray
        (2,) Array with coordinates of vertex `b` of the triangle
    c : ndarray
        (2,) Array with coordinates of vertex `c` of the triangle

    Returns
    -------
    r : float
        Circumcircle of the triangle

    Notes
    -----
    Source for equations:
    > https://www.mathopenref.com/trianglecircumcircle.html
    [Last accessed July 11th. 2018]

    Examples
    --------
    >>> a = np.array([0, 0])
    >>> b = np.array([0.5, 0])
    >>> c = np.array([0.25, 0.25])
    >>> r = r_circumcircle_triangle_single(a, b, c)
    >>> r
    0.2500000000000001
    """
    ab = nb_dist(a, b)
    bc = nb_dist(b, c)
    ca = nb_dist(c, a)

    num = ab * bc * ca
    den = np.sqrt((ab + bc + ca) * (bc + ca - ab) * (ca + ab - bc) * (ab + bc - ca))
    if den == 0:
        return np.array([ab, bc, ca]).max() / 2.0
    else:
        return num / den


@jit(nopython=True)
def r_circumcircle_triangle(a_s, b_s, c_s):
    """Computation of circumcircles for a series of triangles

    Parameters
    ----------
    a_s : ndarray
        (N, 2) array with coordinates of vertices `a` of the triangles
    b_s : ndarray
        (N, 2) array with coordinates of vertices `b` of the triangles
    c_s : ndarray
        (N, 2) array with coordinates of vertices `c` of the triangles

    Returns
    -------
    radii : ndarray
        (N,) array with circumcircles for every triangle

    Examples
    --------
    >>> a_s = np.array([[0, 0], [2, 1], [3, 2]])
    >>> b_s = np.array([[1, 0], [5, 1], [2, 4]])
    >>> c_s = np.array([[0, 7], [1, 3], [4, 2]])
    >>> rs = r_circumcircle_triangle(a_s, b_s, c_s)
    >>> rs
    array([3.53553391, 2.5       , 1.58113883])
    """
    len_a = len(a_s)
    r2 = np.zeros((len_a,))
    for i in range(len_a):
        r2[i] = r_circumcircle_triangle_single(a_s[i], b_s[i], c_s[i])
    return r2


@jit(nopython=True)
def get_faces(triangle):
    """Extract faces from a single triangle

    Parameters
    ----------
    triangles : ndarray
        (3,) array with the vertex indices for a triangle

    Returns
    -------
    faces : ndarray
        (3, 2) array with a row for each face containing the indices of the two
        points that make up the face

    Examples
    --------
    >>> triangle = np.array([3, 1, 4], dtype=np.int32)
    >>> faces = get_faces(triangle)
    >>> faces
    array([[3., 1.],
           [1., 4.],
           [4., 3.]])
    """
    faces = np.zeros((3, 2))
    for i, (i0, i1) in enumerate([(0, 1), (1, 2), (2, 0)]):
        faces[i] = triangle[i0], triangle[i1]
    return faces


@jit(nopython=True)
def build_faces(faces, triangles_is, num_triangles, num_faces_single):
    """Build facing triangles

    Parameters
    ----------
    faces : ndarray
        (num_triangles * num_faces_single, 2) array of zeroes in int form
    triangles_is : ndarray
        (D, 3) array, where D is the number of Delaunay triangles, with the
        vertex indices for each triangle
    num_triangles : int
        Number of triangles
    num_faces_single : int
        Number of faces a triangle has (i.e. 3)

    Returns
    -------
    faces : ndarray
        Two dimensional array with a row for every facing segment containing
        the indices of the coordinate points

    Examples
    --------
    >>> import scipy.spatial as spat
    >>> pts = np.array([[0, 1], [3, 5], [4, 1], [6, 7], [9, 3]])
    >>> triangulation = spat.Delaunay(pts)
    >>> triangulation.simplices
    array([[3, 1, 4],
           [1, 2, 4],
           [2, 1, 0]], dtype=int32)
    >>> num_faces_single = 3
    >>> num_triangles = triangulation.simplices.shape[0]
    >>> num_faces = num_triangles * num_faces_single
    >>> faces = np.zeros((num_faces, 2), dtype=np.int_)
    >>> mask = np.ones((num_faces,), dtype=np.bool_)
    >>> faces = build_faces(
    ...     faces, triangulation.simplices, num_triangles, num_faces_single
    ... )
    >>> faces
    array([[3, 1],
           [1, 4],
           [4, 3],
           [1, 2],
           [2, 4],
           [4, 1],
           [2, 1],
           [1, 0],
           [0, 2]])
    """
    for i in range(num_triangles):
        from_i = num_faces_single * i
        to_i = num_faces_single * (i + 1)
        faces[from_i:to_i] = get_faces(triangles_is[i])
    return faces


@jit(nopython=True)
def nb_mask_faces(mask, faces):
    """Run over each row in `faces`, if the face in the following row is the
    same, then mark both as False on `mask`

    Parameters
    ----------
    mask : ndarray
        One-dimensional boolean array set to True with as many observations as
        rows in `faces`
    faces : ndarray
        Sorted sequence of faces for all triangles (ie. triangles split by each
        segment)

    Returns
    -------
    masked : ndarray
         Sequence of outward-facing faces

    Examples
    --------
    >>> import numpy as np
    >>> faces = np.array(
    ...     [
    ...         [0, 1], [0, 2], [1, 2], [1, 2], [1, 3], [1, 4], [1, 4], [2, 4], [3, 4]
    ...     ]
    ... )
    >>> mask = np.ones((faces.shape[0], ), dtype=np.bool_)
    >>> masked = nb_mask_faces(mask, faces)
    >>> masked
    array([[0, 1],
           [0, 2],
           [1, 3],
           [2, 4],
           [3, 4]])
    """
    for k in range(faces.shape[0] - 1):
        if mask[k] and np.all(faces[k] == faces[k + 1]):
            mask[k] = False
            mask[k + 1] = False
    return faces[mask]


def get_single_faces(triangles_is):
    """Extract outward facing edges from collection of triangles

    Parameters
    ----------
    triangles_is : ndarray
        (D, 3) array, where D is the number of Delaunay triangles, with the
        vertex indices for each triangle

    Returns
    -------
    single_faces : ndarray

    Examples
    --------
    >>> import scipy.spatial as spat
    >>> pts = np.array([[0, 1], [3, 5], [4, 1], [6, 7], [9, 3]])
    >>> alpha = 0.33
    >>> triangulation = spat.Delaunay(pts)
    >>> triangulation.simplices
    array([[3, 1, 4],
           [1, 2, 4],
           [2, 1, 0]], dtype=int32)
    >>> get_single_faces(triangulation.simplices)
    array([[0, 1],
           [0, 2],
           [1, 3],
           [2, 4],
           [3, 4]])
    """
    num_faces_single = 3
    num_triangles = triangles_is.shape[0]
    num_faces = num_triangles * num_faces_single
    faces = np.zeros((num_faces, 2), dtype=np.int_)
    mask = np.ones((num_faces,), dtype=np.bool_)

    faces = build_faces(faces, triangles_is, num_triangles, num_faces_single)

    orderlist = [f"x{i}" for i in range(faces.shape[1])]
    dtype_list = [(el, faces.dtype.str) for el in orderlist]
    # Arranging each face so smallest vertex is first
    faces.sort(axis=1)
    # Arranging faces in ascending way
    faces.view(dtype_list).sort(axis=0)
    # Masking
    single_faces = nb_mask_faces(mask, faces)
    return single_faces


@requires("geopandas", "shapely")
def _alpha_geoms(alpha, triangles, radii, xys):
    """Generate alpha-shape polygon(s) from `alpha` value, vertices of
    `triangles`, the `radii` for all points, and the points themselves

    Parameters
    ----------
    alpha : float
        Alpha value to delineate the alpha-shape
    triangles : ndarray
         (D, 3) array, where D is the number of Delaunay triangles, with the
         vertex indices for each triangle
    radii : ndarray
        (N,) array with circumcircles for every triangle
    xys : ndarray
        (N, 2) array with one point per row and coordinates structured as X and Y

    Returns
    -------
    geoms : GeoSeries
        Polygon(s) resulting from the alpha shape algorithm, in a GeoSeries.
        The output is a GeoSeries even if only a single polygon is returned. There is
        no CRS included in the returned GeoSeries.

    Examples
    --------
    >>> import scipy.spatial as spat
    >>> pts = np.array([[0, 1], [3, 5], [4, 1], [6, 7], [9, 3]])
    >>> alpha = 0.33
    >>> triangulation = spat.Delaunay(pts)
    >>> triangles = pts[triangulation.simplices]
    >>> triangles
    array([[[6, 7],
            [3, 5],
            [9, 3]],
    <BLANKLINE>
           [[3, 5],
            [4, 1],
            [9, 3]],
    <BLANKLINE>
           [[4, 1],
            [3, 5],
            [0, 1]]])
    >>> a_pts = triangles[:, 0, :]
    >>> b_pts = triangles[:, 1, :]
    >>> c_pts = triangles[:, 2, :]
    >>> radii = r_circumcircle_triangle(a_pts, b_pts, c_pts)
    >>> geoms = alpha_geoms(alpha, triangulation.simplices, radii, pts)
    >>> geoms
    0    POLYGON ((0.00000 1.00000, 3.00000 5.00000, 4....
    dtype: geometry
    """
    from geopandas import GeoSeries
    from shapely.geometry import LineString
    from shapely.ops import polygonize

    triangles_reduced = triangles[radii < 1 / alpha]
    outer_triangulation = get_single_faces(triangles_reduced)
    face_pts = xys[outer_triangulation]
    geoms = GeoSeries(list(polygonize(list(map(LineString, face_pts)))))
    return geoms


@requires("geopandas", "shapely")
def alpha_shape(xys, alpha):
    """Alpha-shape delineation (Edelsbrunner, Kirkpatrick & Seidel, 1983)
    from a collection of points

    Parameters
    ----------
    xys : ndarray
        (N, 2) array with one point per row and coordinates structured as X and Y
    alpha : float
        Alpha value to delineate the alpha-shape

    Returns
    -------
    shapes : GeoSeries
         Polygon(s) resulting from the alpha shape algorithm. The GeoSeries
         object remains so even if only a single polygon is returned. There is
         no CRS included in the object. Note that the returned shape(s) may
         have holes, as per the definition of the shape in Edselbrunner et al. (1983)

    Examples
    --------
    >>> pts = np.array([[0, 1], [3, 5], [4, 1], [6, 7], [9, 3]])
    >>> alpha = 0.1
    >>> poly = alpha_shape(pts, alpha)
    >>> poly
    0    POLYGON ((0.00000 1.00000, 3.00000 5.00000, 6....
    dtype: geometry
    >>> poly.centroid
    0    POINT (4.69048 3.45238)
    dtype: geometry

    References
    ----------
    Edelsbrunner, H., Kirkpatrick, D., & Seidel, R. (1983). On the shape of
        a set of points in the plane. IEEE Transactions on information theory,
        29(4), 551-559.
    """
    if not HAS_JIT:
        warn(NUMBA_WARN, stacklevel=2)
    if xys.shape[0] < 4:
        from shapely import geometry as geom
        from shapely import ops

        return ops.unary_union([geom.Point(xy) for xy in xys]).convex_hull.buffer(0)
    triangulation = spat.Delaunay(xys)
    triangles = xys[triangulation.simplices]
    a_pts = triangles[:, 0, :]
    b_pts = triangles[:, 1, :]
    c_pts = triangles[:, 2, :]
    radii = r_circumcircle_triangle(a_pts, b_pts, c_pts)
    del triangles, a_pts, b_pts, c_pts
    geoms = _alpha_geoms(alpha, triangulation.simplices, radii, xys)
    geoms = _filter_holes(geoms, xys)
    return geoms


def _valid_hull(geoms, points):
    """Sanity check within ``alpha_shape_auto()`` to verify the generated alpha
    shape actually contains the original set of points (xys).

    Parameters
    ----------
    geoms : GeoSeries
        See alpha_geoms()
    points : list
        xys parameter cast as shapely.geometry.Point objects

    Returns
    -------
    flag : bool
        Valid hull for alpha shape [True] or not [False]
    """
    # if there is not exactly one polygon
    if geoms.shape[0] != 1:
        return False
    # if any (xys) points do not intersect the polygon
    if HAS_SHAPELY:
        return shapely.intersects(geoms[0], points).all()
    else:
        return all(point.intersects(geoms[0]) for point in points)


@requires("geopandas", "shapely")
def alpha_shape_auto(
    xys, step=1, verbose=False, return_radius=False, return_circles=False
):
    """Computation of alpha-shape delineation with automated selection of alpha.

    This method uses the algorithm proposed by  Edelsbrunner, Kirkpatrick &
    Seidel (1983) to return the tightest polygon that contains all points in
    `xys`. The algorithm ranks every point based on its radius and iterates
    over each point, checking whether the maximum alpha that would keep the
    point and all the other ones in the set with smaller radii results in a
    single polygon. If that is the case, it moves to the next point;
    otherwise, it retains the previous alpha value and returns the polygon
    as `shapely` geometry. Note that this geometry may have holes.

    Parameters
    ----------

    xys : ndarray
        Nx2 array with one point per row and coordinates structured as X and Y
    step : int
        [Optional. Default=1] Number of points in `xys` to jump ahead after
        checking whether the largest possible alpha that includes the point and
        all the other ones with smaller radii
    verbose : Boolean
        [Optional. Default=False] If True, it prints alpha values
        being tried at every step.

    Returns
    -------
    poly : shapely.Polygon
         Tightest alpha-shape polygon containing all points in `xys`

    Examples
    --------
    >>> pts = np.array([[0, 1], [3, 5], [4, 1], [6, 7], [9, 3]])
    >>> poly = alpha_shape_auto(pts)
    >>> poly.bounds
    (0.0, 1.0, 9.0, 7.0)
    >>> poly.centroid.x, poly.centroid.y
    (4.690476190476191, 3.4523809523809526)

    References
    ----------
    Edelsbrunner, H., Kirkpatrick, D., & Seidel, R. (1983). On the shape of
        a set of points in the plane. IEEE Transactions on information theory,
        29(4), 551-559.

    """
    if not HAS_JIT:
        warn(NUMBA_WARN, stacklevel=2)
    from shapely import geometry as geom

    if return_circles:
        return_radius = True
    if xys.shape[0] < 4:
        from shapely import ops

        if xys.shape[0] == 3:
            multipoint = ops.cascaded_union([geom.Point(xy) for xy in xys])
            alpha_shape = multipoint.convex_hull.buffer(0)
        else:
            alpha_shape = geom.Polygon([])
        if xys.shape[0] == 1:
            if return_radius:
                if return_circles:
                    out = [alpha_shape, 0, alpha_shape]
                return alpha_shape, 0
            return alpha_shape
        elif xys.shape[0] == 2:
            if return_radius:
                r = spat.distance.euclidean(xys[0], xys[1]) / 2
                if return_circles:
                    circle = _construct_centers(xys[0], xys[1], r)
                    return [alpha_shape, r, circle]
                return [alpha_shape, r]
            return alpha_shape
        elif return_radius:  # this handles xys.shape[0] == 3
            radius = r_circumcircle_triangle_single(xys[0], xys[1], xys[2])
            if return_circles:
                circles = construct_bounding_circles(alpha_shape, radius)
                return [alpha_shape, radius, circles]
            return [alpha_shape, radius]
        return alpha_shape
    triangulation = spat.Delaunay(xys)
    triangles = xys[triangulation.simplices]
    a_pts = triangles[:, 0, :]
    b_pts = triangles[:, 1, :]
    c_pts = triangles[:, 2, :]
    radii = r_circumcircle_triangle(a_pts, b_pts, c_pts)
    radii[np.isnan(radii)] = 0  # "Line" triangles to be kept for sure
    del triangles, a_pts, b_pts, c_pts
    radii_sorted_i = radii.argsort()
    triangles = triangulation.simplices[radii_sorted_i][::-1]
    radii = radii[radii_sorted_i][::-1]
    geoms_prev = _alpha_geoms((1 / radii.max()) - EPS, triangles, radii, xys)
    points = shapely.points(xys) if HAS_SHAPELY else [geom.Point(pnt) for pnt in xys]
    if verbose:
        print("Step set to %i" % step)
    for i in range(0, len(radii), step):
        radi = radii[i]
        alpha = (1 / radi) - EPS
        if verbose:
            print(f"{(i + 1) / radii.shape[0]:.2f}% | Trying a = {alpha:f}")
        geoms = _alpha_geoms(alpha, triangles, radii, xys)
        if _valid_hull(geoms, points):
            geoms_prev = geoms
            radi_prev = radi
        else:
            break
    if verbose:
        print(geoms_prev.shape)
    if return_radius:
        out = [geoms_prev[0], radi_prev]
        if return_circles:
            out.append(construct_bounding_circles(out[0], radi_prev))
        return out
    # Return a shapely polygon
    return geoms_prev[0]


def construct_bounding_circles(alpha_shape, radius):
    """Construct the bounding circles for an alpha shape, given the radius
    computed from the `alpha_shape_auto` method.

    Parameters
    ----------
    alpha_shape : shapely.Polygon
        An alpha-hull with the input radius.
    radius : float
        The radius of the input alpha_shape.

    Returns
    -------
    center : numpy.ndarray of shape (n,2)
        The centers of the circles defining the alpha_shape.
    """
    coordinates = list(alpha_shape.boundary.coords)
    n_coordinates = len(coordinates)
    centers = []
    for i in range(n_coordinates - 1):
        a, b = coordinates[i], coordinates[i + 1]
        centers.append(_construct_centers(a, b, radius))
    return centers


@jit(nopython=True)
def _construct_centers(a, b, radius):
    midpoint_x = (a[0] + b[0]) * 0.5
    midpoint_y = (a[1] + b[1]) * 0.5
    d = ((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2) ** 0.5
    if b[0] - a[0] == 0:
        m = np.inf
        axis_rotation = np.pi / 2
    else:
        m = (b[1] - a[1]) / (b[0] - a[0])
        axis_rotation = np.arctan(m)
    # altitude is perpendicular bisector of AB
    interior_angle = np.arccos(0.5 * d / radius)
    chord = np.sin(interior_angle) * radius

    dx = chord * np.sin(axis_rotation)
    dy = chord * np.cos(axis_rotation)

    up_x = midpoint_x - dx
    up_y = midpoint_y + dy
    down_x = midpoint_x + dx
    down_y = midpoint_y - dy

    # sign gives us direction of point, since
    # shapely shapes are clockwise-defined
    sign = np.sign((b[0] - a[0]) * (up_y - a[1]) - (b[1] - a[1]) * (up_x - a[0]))
    if sign == 1:
        return up_x, up_y
    else:
        return down_x, down_y


def _filter_holes(geoms, points):  # noqa ARG001
    """
    Filter hole polygons using a computational geometry solution
    """
    import geopandas

    if (geoms.interiors.apply(len) > 0).any():
        from shapely.geometry import Polygon

        # Extract the "shell", or outer ring of the polygon.
        shells = geoms.exterior.apply(Polygon)
        # Compute which original geometries are within each shell, self-inclusive
        if Version(geopandas.__version__) >= Version("0.13"):
            inside, outside = shells.sindex.query(geoms, predicate="within")
        else:
            inside, outside = shells.sindex.query_bulk(geoms, predicate="within")

        # Now, create the sparse matrix relating the inner geom (rows)
        # to the outer shell (cols) and take the sum.
        # A z-order of 1 means the polygon is only inside if its own exterior.
        #   This means it's not a hole.
        # A z-order of 2 means the polygon is inside of exactly one other exterior.
        #   Because the hull generation method is restricted to be planar, this
        #   means the polygon is a hole.
        # In general, an even z-order means that the polygon is always exactly
        #   matched to one exterior, plus some number of intermediate
        #   exterior-hole pairs. Therefore, the polygon is a hole.
        # In general, an odd z-order means that there is an uneven number of exteriors.
        #   This means the polygon is not a hole.
        zorder = sparse.csc_matrix((np.ones_like(inside), (inside, outside))).sum(
            axis=1
        )
        zorder = np.asarray(zorder).flatten()
        # Keep only the odd z-orders
        to_include = (zorder % 2).astype(bool)
        geoms = geoms[to_include]
    return geoms


if __name__ == "__main__":
    import time

    import geopandas as gpd
    import matplotlib.pyplot as plt

    plt.close("all")
    xys = np.random.random((1000, 2))
    t0 = time.time()
    geoms = alpha_shape_auto(xys, 1)
    t1 = time.time()
    print("%.2f Seconds to run algorithm" % (t1 - t0))
    f, ax = plt.subplots(1)
    gpd.GeoDataFrame({"geometry": [geoms]}).plot(ax=ax, color="orange", alpha=0.5)
    ax.scatter(xys[:, 0], xys[:, 1], s=0.1)
    plt.show()



"""
FileIO: Module for reading and writing various file types in a Pythonic way.
This module should not be used directly, instead...

```
import pysal.core.FileIO as FileIO
```

Readers and Writers will mimic python file objects.
    .seek(n) seeks to the n'th object
    .read(n) reads n objects, default == all
    .next() reads the next object

"""


# ruff: noqa: ARG002, N801, N802, N803, N806, SIM115

__author__ = "Charles R Schmidt <schmidtc@gmail.com>"

__all__ = ["FileIO"]

import os.path
from warnings import warn

from ..common import MISSINGVALUE


class FileIO_MetaCls(type):
    """This meta class is instantiated when the class is first defined. All
    subclasses of `FileIO` also inherit this meta class, which registers
    their abilities with the FileIO registry. Subclasses must contain
    ``FORMATS`` and ``MODES`` (both are ``type(list)``).

    Raises
    ------
    TypeError
        FileIO subclasses must have ``FORMATS`` and ``MODES`` defined.
    """

    def __new__(cls, name, bases, dict_):
        cls = type.__new__(cls, name, bases, dict_)

        if name != "FileIO" and name != "DataTable":
            if "FORMATS" in dict_ and "MODES" in dict_:
                # msg = "Registering %s with FileIO.\n\tFormats: %r\n\tModes: %r"
                # msg = msg % (name, dict["FORMATS"], dict["MODES"])
                FileIO._register(cls, dict_["FORMATS"], dict_["MODES"])
            else:
                raise TypeError(
                    "FileIO subclasses must have 'FORMATS' and 'MODES' defined."
                )

        return cls


class FileIO(metaclass=FileIO_MetaCls):  # should be a type?
    """Metaclass for supporting spatial data file read and write.

    How this works:

    ``FileIO.open(\\*args) == FileIO(\\*args)``

    When creating a new instance of `FileIO` the ``.__new__`` method intercepts.
    ``.__new__`` parses the filename to determine the ``fileType``. Next,
    ``.__registry`` and checked for that type. Each type supports one or more modes
    (``['r', 'w', 'a', etc.]``). If we support the type and mode, an instance of the
    appropriate handler is created and returned. All handlers must inherit from this
    class, and by doing so are automatically added to the ``.__registry`` and are
    forced to conform to the prescribed API. The metaclass takes care of the
    registration by parsing the class definition. It doesn't make much sense to
    treat weights in the same way as shapefiles and dbfs, so...

    * ... for now we'll just return an instance of `W` on ``mode='r'``.
    * ... on ``mode='w'``, ``.write`` will expect an instance of `W`.
    """

    __registry = {}  # {'shp':{'r':[OGRshpReader,pysalShpReader]}}

    def __new__(cls, dataPath="", mode="r", dataFormat=None):
        """Intercepts the instantiation of ``FileIO`` and dispatches
        to the correct handler. If no suitable handler is found a
        python file object is returned.
        """

        if cls is FileIO:
            try:
                newCls = object.__new__(
                    cls.__registry[cls.getType(dataPath, mode, dataFormat)][mode][0]
                )
            except KeyError:
                return open(dataPath, mode)
            return newCls
        else:
            return object.__new__(cls)

    @staticmethod
    def getType(dataPath: str, mode: str, dataFormat=None) -> str:
        """Parse the ``dataPath`` and return the data type."""

        if dataFormat:
            ext = dataFormat
        else:
            ext = os.path.splitext(dataPath)[1]
            ext = ext.replace(".", "")
            ext = ext.lower()
        if ext == "txt":
            with open(dataPath, mode) as f:
                l1 = f.readline()
                l2 = f.readline()
                try:
                    n, k = l1.split(",")
                    n, k = int(n), int(k)
                    fields = l2.split(",")
                    assert len(fields) == k
                    return "geoda_txt"
                except AssertionError:
                    return ext

        return ext

    @classmethod
    def _register(cls, parser, formats, modes):
        """This method is called automatically via the Metaclass of `FileIO` subclasses
        This should be private, but that hides it from the Metaclass.
        """

        assert cls is FileIO

        for format_ in formats:
            if format_ not in cls.__registry:
                cls.__registry[format_] = {}
            for mode in modes:
                if mode not in cls.__registry[format_]:
                    cls.__registry[format_][mode] = []
                cls.__registry[format_][mode].append(parser)
        # cls.check()

    @classmethod
    def check(cls):
        """Prints the contents of the registry."""

        print("PySAL File I/O understands the following file extensions:")

        for key, val in list(cls.__registry.items()):
            print(f"Ext: '.{key}', Modes: {list(val.keys())!r}")

    @classmethod
    def open(cls, *args, **kwargs):  # noqa A001
        """Alias for ``FileIO()``."""

        return cls(*args, **kwargs)

    class _By_Row:
        def __init__(self, parent):
            self.p = parent

        def __repr__(self) -> str:
            if not self.p.ids:
                return "keys: range(0,n)"
            else:
                return "keys: " + list(self.p.ids.keys()).__repr__()

        def __getitem__(self, key) -> list | str:
            if isinstance(key, list):
                r = []
                if self.p.ids:
                    for k in key:
                        r.append(self.p.get(self.p.ids[k]))
                else:
                    for k in key:
                        r.append(self.p.get(k))
                return r
            if self.p.ids:
                return self.p.get(self.p.ids[key])
            else:
                return self.p.get(key)

        __call__ = __getitem__

    def __init__(self, dataPath="", mode="r", dataFormat=None):
        self.dataPath = dataPath
        self.dataObj = ""
        self.mode = mode
        # pos Should ALWAYS be in the range 0,...,n
        # for custom IDs set the ids property.
        self.pos = 0
        self.__ids = None  # {'id':n}
        self.__rIds = None
        self.closed = False
        self._spec = []
        self.header = []

    def __getitem__(self, key):
        return self.by_row.__getitem__(key)

    @property
    def by_row(self):
        return self._By_Row(self)

    def __getIds(self):
        return self.__ids

    def __setIds(self, ids: list | (dict | None)):
        """Property method for ``.ids``. Takes a list of ids and maps then
        to a 0-based index. Need to provide a method to set ID's based on
        a ``fieldName`` preferably without reading the whole file.

        Raises
        ------
        AssertionError
            Raised when IDs are not unique.
        """

        if isinstance(ids, list):
            try:
                assert len(ids) == len(set(ids))
            except AssertionError:
                raise KeyError("IDs must be unique.") from None
            # keys: ID values: i
            self.__ids = {}
            # keys: i values: ID
            self.__rIds = {}
            for i, id_ in enumerate(ids):
                self.__ids[id] = i
                self.__rIds[i] = id_
        elif isinstance(ids, dict):
            self.__ids = ids
            self.__rIds = {}
            for id_, n in list(ids.items()):
                self.__rIds[n] = id_
        elif not ids:
            self.__ids = None
            self.__rIds = None

    ids = property(fget=__getIds, fset=__setIds)

    @property
    def rIds(self) -> dict | None:
        return self.__rIds

    def __iter__(self):
        self.seek(0)
        return self

    @staticmethod
    def _complain_ifclosed(closed):
        """From `StringIO`.

        Raises
        ------
        ValueError
            Raised when a file is already closed.
        """
        if closed:
            raise ValueError("I/O operation on closed file.")

    def cast(self, key, typ):
        """Cast ``key`` as ``typ``.

        Raises
        ------
        TypeError
            Raised when a cast object in not callable.
        KeyError
            Raised when a key is not present.
        """
        if key in self.header:
            if not self._spec:
                self._spec = [lambda x: x for k in self.header]
            if typ is None:
                self._spec[self.header.index(key)] = lambda x: x
            else:
                try:
                    assert callable(typ)
                    self._spec[self.header.index(key)] = typ
                except AssertionError:
                    raise TypeError("Cast objects must be callable.") from None
        else:
            raise KeyError("%s" % key)

    def _cast(self, row) -> list:
        """

        Raises
        ------
        ValueError
            Raised when a value could not be cast a particular type.
        """
        if self._spec and row:
            try:
                return [f(v) for f, v in zip(self._spec, row, strict=True)]
            except ValueError:
                r = []
                for f, v in zip(self._spec, row, strict=True):
                    try:
                        if not v and f != str:
                            raise ValueError
                        r.append(f(v))
                    except ValueError:
                        msg = "Value '%r' could not be cast to %s, "
                        msg += "value set to MISSINGVALUE."
                        msg = msg % (v, str(f))
                        warn(msg, RuntimeWarning, stacklevel=2)
                        r.append(MISSINGVALUE)
                return r

        else:
            return row

    def __next__(self) -> list:
        """A `FileIO` object is its own iterator, see `StringIO`.

        Raises
        ------
        StopIteration
            Raised at the EOF.
        """

        self._complain_ifclosed(self.closed)
        r = self.__read()
        if r is None:
            raise StopIteration

        return r

    def close(self):
        """Subclasses should clean themselves up and then call this method."""

        if not self.closed:
            self.closed = True
            del self.dataObj, self.pos

    def get(self, n: int) -> list:
        """Seeks the file to ``n`` and returns ``n``. If ``.ids`` is set
        ``n`` should be an id, else, ``n`` should be an offset.
        """

        prev_pos = self.tell()
        self.seek(n)
        obj = self.__read()
        self.seek(prev_pos)

        return obj

    def seek(self, n: int):
        """Seek the `FileObj` to the beginning of the ``n``'th record.
        If IDs are set, seeks to the beginning of the record at ID, ``n``.
        """

        self._complain_ifclosed(self.closed)
        self.pos = n

    def tell(self) -> int:
        """Return ID (or offset) of next object."""

        self._complain_ifclosed(self.closed)

        return self.pos

    def read(self, n=-1) -> list | None:
        """Read at most ``n`` objects, less if read hits EOF.
        If size is negative or omitted read all objects until EOF.
        Returns ``None`` if EOF is reached before any objects.

        Raises
        ------
        StopIteration
            Raised at the EOF.
        """

        self._complain_ifclosed(self.closed)

        if n < 0:
            # return list(self)
            result = []
            while 1:
                try:
                    result.append(self.__read())
                except StopIteration:
                    break
            return result
        elif n == 0:
            return None
        else:
            result = []
            for _i in range(0, n):
                try:
                    result.append(self.__read())
                except StopIteration:
                    break
            return result

    def __read(self) -> list:
        """Gets one row from the file handler, and if necessary casts it's objects.

        Raises
        ------
        StopIteration
            Raised at the EOF.
        """

        row = self._read()
        if row is None:
            raise StopIteration
        row = self._cast(row)

        return row

    def _read(self):
        """Must be implemented by subclasses that support 'r' subclasses.
        Should increment ``.pos`` and redefine this doc string.

        Raises
        ------
        NotImplementedError
        """

        self._complain_ifclosed(self.closed)
        raise NotImplementedError

    def truncate(self, size=None):
        """Should be implemented by subclasses and redefine this doc string.

        Raises
        ------
        NotImplementedError
        """

        self._complain_ifclosed(self.closed)
        raise NotImplementedError

    def write(self, obj):
        """Must be implemented by subclasses that support 'w' subclasses
        Should increment ``.pos``. Subclasses should also check if ``obj``
        is an instance of type(list) and redefine this doc string.

        Raises
        ------
        NotImplementedError
        """

        self._complain_ifclosed(self.closed)
        "Write obj to dataObj"
        raise NotImplementedError

    def flush(self):
        """

        Raises
        ------
        NotImplementedError
        """

        self._complain_ifclosed(self.closed)
        raise NotImplementedError



import numpy as np

from libpysal import graph
from libpysal.graph._spatial_lag import _lag_spatial
from libpysal.weights import lat2W


class TestLag:
    def setup_method(self):
        self.neighbors = {
            "a": ["b"],
            "b": ["c", "a"],
            "c": ["b"],
            "d": [],
        }
        self.weights = {"a": [1.0], "b": [1.0, 1.0], "c": [1.0], "d": []}
        self.g = graph.Graph.from_dicts(self.neighbors, self.weights)
        self.y = np.array([0, 1, 2, 3])

    def test_lag_spatial(self):
        yl = _lag_spatial(self.g, self.y)
        np.testing.assert_array_almost_equal(yl, [1.0, 2.0, 1.0, 0])
        g = graph.Graph.from_W(lat2W(3, 3))
        y = np.arange(9)
        yl = _lag_spatial(g, y)
        ylc = np.array([4.0, 6.0, 6.0, 10.0, 16.0, 14.0, 10.0, 18.0, 12.0])
        np.testing.assert_array_almost_equal(yl, ylc)
        g_row = g.transform("r")
        yl = _lag_spatial(g_row, y)
        ylc = np.array([2.0, 2.0, 3.0, 3.33333333, 4.0, 4.66666667, 5.0, 6.0, 6.0])
        np.testing.assert_array_almost_equal(yl, ylc)



import geodatasets
import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_almost_equal
from scipy.sparse import csr_matrix
from shapely import get_coordinates

from libpysal import graph

TRIANGULATIONS = ["delaunay", "gabriel", "relative_neighborhood", "voronoi"]

"""
This file tests Graph initialisation from various build_* constructors. The correctness
of the underlying data shall be tested in respective constructor test suites.
"""


class TestContiguity:
    def setup_method(self):
        self.gdf = gpd.read_file(geodatasets.get_path("nybb"))
        self.gdf_str = self.gdf.set_index("BoroName")

    def test_vertex_intids(self):
        g = graph.Graph.build_contiguity(self.gdf)

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_vertex_strids(self):
        g = graph.Graph.build_contiguity(self.gdf_str)

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_rook_intids(self):
        g = graph.Graph.build_contiguity(self.gdf, strict=True)

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_rook_strids(self):
        g = graph.Graph.build_contiguity(self.gdf_str, strict=True)

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_queen_intids(self):
        g = graph.Graph.build_contiguity(self.gdf, rook=False, strict=True)

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_queen_strids(self):
        g = graph.Graph.build_contiguity(self.gdf_str, rook=False, strict=True)

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_vertex_intids_perimeter(self):
        g = graph.Graph.build_contiguity(self.gdf, by_perimeter=True)

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_vertex_strid_perimeters(self):
        g = graph.Graph.build_contiguity(self.gdf_str, by_perimeter=True)

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_rook_intids_perimeter(self):
        g = graph.Graph.build_contiguity(self.gdf, strict=True, by_perimeter=True)

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_rook_strid_perimeters(self):
        g = graph.Graph.build_contiguity(self.gdf_str, strict=True, by_perimeter=True)

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_queen_intid_perimeters(self):
        g = graph.Graph.build_contiguity(
            self.gdf, rook=False, strict=True, by_perimeter=True
        )

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_queen_strids_perimeter(self):
        g = graph.Graph.build_contiguity(
            self.gdf_str, rook=False, strict=True, by_perimeter=True
        )

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_block_contiguity(self):
        regimes = ["n", "n", "s", "s", "e", "e", "w", "w", "e", "l"]
        g = graph.Graph.build_block_contiguity(regimes)

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

        regimes = pd.Series(
            regimes, index=["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"]
        )
        g = graph.Graph.build_block_contiguity(regimes)
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_fuzzy_contiguity_intids(self):
        g = graph.Graph.build_fuzzy_contiguity(self.gdf)

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_fuzzy_contiguity_strids(self):
        g = graph.Graph.build_fuzzy_contiguity(self.gdf_str)

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)


class TestTriangulation:
    def setup_method(self):
        gdf = gpd.read_file(geodatasets.get_path("geoda liquor_stores")).explode(
            ignore_index=True
        )
        self.gdf = gdf[~gdf.geometry.duplicated()]
        self.gdf_str = self.gdf.set_index("placeid")

    @pytest.mark.parametrize("method", TRIANGULATIONS)
    def test_triangulation_intids(self, method):
        g = graph.Graph.build_triangulation(self.gdf, method)

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    @pytest.mark.parametrize("method", TRIANGULATIONS)
    def test_triangulation_strids(self, method):
        g = graph.Graph.build_triangulation(self.gdf_str, method)

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    @pytest.mark.parametrize("method", TRIANGULATIONS)
    def test_triangulation_intids_kernel(self, method):
        g = graph.Graph.build_triangulation(
            self.gdf, method, kernel="parabolic", bandwidth=7500
        )
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    @pytest.mark.parametrize("method", TRIANGULATIONS)
    def test_triangulation_strids_kernel(self, method):
        g = graph.Graph.build_triangulation(
            self.gdf_str, method, kernel="parabolic", bandwidth=7500
        )

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_invalid_method(self):
        with pytest.raises(ValueError, match="Method 'invalid' is not supported"):
            graph.Graph.build_triangulation(
                self.gdf, method="invalid", kernel="parabolic", bandwidth=7500
            )


class TestKernel:
    def setup_method(self):
        self.gdf = gpd.read_file(geodatasets.get_path("geoda liquor_stores")).explode(
            ignore_index=True
        )
        self.gdf_str = self.gdf.set_index("placeid")

    def test_kernel_precompute(self):
        sklearn = pytest.importorskip("sklearn")
        df = gpd.read_file(geodatasets.get_path("nybb"))
        df = df.to_crs(df.estimate_utm_crs())
        distmat = csr_matrix(
            sklearn.metrics.pairwise.euclidean_distances(get_coordinates(df.centroid))
        )
        g = graph.Graph.build_kernel(distmat, metric="precomputed")
        expected = np.array(
            [
                0.07131664,
                0.14998932,
                0.09804811,
                0.0402638,
                0.07131664,
                0.18556845,
                0.17529176,
                0.16394507,
                0.14998932,
                0.18556845,
                0.17495794,
                0.11561449,
                0.09804811,
                0.17529176,
                0.17495794,
                0.19116432,
                0.0402638,
                0.16394507,
                0.11561449,
                0.19116432,
            ]
        )

        assert_array_almost_equal(g.adjacency.values, expected, 3)

    def test_kernel_intids(self):
        g = graph.Graph.build_kernel(self.gdf)

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_kernel_strids(self):
        g = graph.Graph.build_kernel(self.gdf_str)

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_knn_intids(self):
        g = graph.Graph.build_knn(self.gdf, k=3)

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_knn_strids(self):
        g = graph.Graph.build_kernel(self.gdf_str, k=3)

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)


class TestDistanceBand:
    def setup_method(self):
        df = gpd.read_file(geodatasets.get_path("nybb"))
        self.gdf = df.set_geometry(df.centroid)
        self.gdf_str = self.gdf.set_index("BoroName")

    def test_distance_band_intids(self):
        g = graph.Graph.build_distance_band(self.gdf, 50000)

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_distance_band_strids(self):
        g = graph.Graph.build_distance_band(self.gdf_str, 50000)

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_distance_band_intids_weighted(self):
        g = graph.Graph.build_distance_band(self.gdf, 50000, binary=False)

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_distance_band_strids_weighted(self):
        g = graph.Graph.build_distance_band(self.gdf_str, 50000, binary=False)

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_distance_band_intids_kernel(self):
        g = graph.Graph.build_distance_band(
            self.gdf, 50000, binary=False, kernel="gaussian"
        )

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_distance_band_strids_kernel(self):
        g = graph.Graph.build_distance_band(
            self.gdf_str, 50000, binary=False, kernel="gaussian"
        )

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)


class TestAdjacency:
    def setup_method(self):
        self.gdf = gpd.read_file(geodatasets.get_path("nybb"))
        self.gdf_str = self.gdf.set_index("BoroName")
        self.expected_adjacency_intid = pd.DataFrame(
            {
                "focal": {
                    0: 0,
                    1: 1,
                    2: 1,
                    3: 1,
                    4: 2,
                    5: 2,
                    6: 3,
                    7: 3,
                    8: 3,
                    9: 4,
                    10: 4,
                },
                "neighbor": {
                    0: 0,
                    1: 2,
                    2: 3,
                    3: 4,
                    4: 1,
                    5: 3,
                    6: 1,
                    7: 2,
                    8: 4,
                    9: 1,
                    10: 3,
                },
                "weight": {
                    0: 0,
                    1: 1,
                    2: 1,
                    3: 1,
                    4: 1,
                    5: 1,
                    6: 1,
                    7: 1,
                    8: 1,
                    9: 1,
                    10: 1,
                },
            }
        )
        self.expected_adjacency_strid = pd.DataFrame(
            {
                "focal": {
                    0: "Staten Island",
                    1: "Queens",
                    2: "Queens",
                    3: "Queens",
                    4: "Brooklyn",
                    5: "Brooklyn",
                    6: "Manhattan",
                    7: "Manhattan",
                    8: "Manhattan",
                    9: "Bronx",
                    10: "Bronx",
                },
                "neighbor": {
                    0: "Staten Island",
                    1: "Brooklyn",
                    2: "Manhattan",
                    3: "Bronx",
                    4: "Queens",
                    5: "Manhattan",
                    6: "Queens",
                    7: "Brooklyn",
                    8: "Bronx",
                    9: "Queens",
                    10: "Manhattan",
                },
                "weight": {
                    0: 0,
                    1: 1,
                    2: 1,
                    3: 1,
                    4: 1,
                    5: 1,
                    6: 1,
                    7: 1,
                    8: 1,
                    9: 1,
                    10: 1,
                },
            }
        )

    def test_adjacency_intids(self):
        g = graph.Graph.from_adjacency(
            self.expected_adjacency_intid,
        )

        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_adjacency_strids(self):
        g = graph.Graph.from_adjacency(
            self.expected_adjacency_strid,
        )

        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["focal"])
        assert pd.api.types.is_string_dtype(g._adjacency.index.dtypes["neighbor"])
        assert pd.api.types.is_numeric_dtype(g._adjacency.dtype)

    def test_adjacency_rename(self):
        adj = self.expected_adjacency_intid
        adj.columns = ["focal", "neighbor", "cost"]  # no longer named weight
        _ = graph.Graph.from_adjacency(adj, weight_col="cost")

    def test_adjacency_wrong(self):
        adj = self.expected_adjacency_intid
        adj.columns = ["focal", "neighbor", "cost"]  # no longer named weight
        with pytest.raises(AssertionError, match='"weight" was given for `weight_col`'):
            graph.Graph.from_adjacency(adj)

    def test_adjacency_match_contiguity(self):
        contiguity = graph.Graph.build_contiguity(self.gdf)
        built = graph.Graph.from_adjacency(self.expected_adjacency_intid)
        assert contiguity == built

        contiguity_str = graph.Graph.build_contiguity(self.gdf_str)
        built_str = graph.Graph.from_adjacency(self.expected_adjacency_strid)
        assert contiguity_str == built_str



"""
For completeness, we need to test a shuffled dataframe
(i.e. always send unsorted data) with:
- numeric ids
- string ids
- point dataframe
- coordinates
- check two kernel functions
- numba/nonumba
"""
import geodatasets
import geopandas
import numpy as np
import pandas as pd
import pytest
import shapely

from libpysal.graph._kernel import _kernel_functions
from libpysal.graph._triangulation import (
    _delaunay,
    _gabriel,
    _relative_neighborhood,
    _voronoi,
)
from libpysal.graph.base import Graph

stores = geopandas.read_file(geodatasets.get_path("geoda liquor_stores")).explode(
    index_parts=False
)
stores_unique = stores.drop_duplicates(subset="geometry")

kernel_functions = [None] + list(_kernel_functions.keys())


def my_kernel(distances, bandwidth):
    output = np.cos(distances / distances.max())
    output[distances < bandwidth] = 0
    return output


kernel_functions.append(my_kernel)

# optimal, small, and larger than largest distance.
bandwidths = [None, "auto", 0.5]

np.random.seed(6301)
# create a 2-d laplace distribution as a "degenerate"
# over-concentrated distribution
lap_coords = np.random.laplace(size=(5, 2))

# create a 2-d cauchy as a "degenerate"
# spatial outlier-y distribution
cau_coords = np.random.standard_cauchy(size=(5, 2))

parametrize_ids = pytest.mark.parametrize(
    "ids", [None, "id", "placeid"], ids=["no id", "id", "placeid"]
)

parametrize_kernelfunctions = pytest.mark.parametrize(
    "kernel", kernel_functions, ids=kernel_functions[:-2] + ["None", "custom callable"]
)
parametrize_bw = pytest.mark.parametrize(
    "bandwidth", bandwidths, ids=["no bandwidth", "auto", "fixed"]
)
parametrize_constructors = pytest.mark.parametrize(
    "constructor",
    [_delaunay, _gabriel, _relative_neighborhood, _voronoi],
    ids=["delaunay", "gabriel", "relhood", "voronoi"],
)

# @parametrize_constructors
# @parametrize_ids
# @parametrize_kernelfunctions``
# @parametrize_bw
# def test_option_combinations(constructor, ids, kernel, bandwidth):
#     """
#     NOTE: This does not check for the *validity* of the output, just
#     the structure of the output.
#     """
#     heads, tails, weights = constructor(
#         stores_unique,
#         ids=stores_unique[ids] if ids is not None else None,
#         kernel=kernel,
#         bandwidth=bandwidth
#         )
#     assert heads.dtype == tails.dtype
#     assert (
#         heads.dtype == stores_unique.get(ids, stores_unique.index).dtype,
#         'ids failed to propagate'
#     )
#     if kernel is None and bandwidth is None:
#         np.testing.assert_array_equal(weights, np.ones_like(heads))
#     assert (
#         set(zip(heads, tails)) == set(zip(tails, heads)),
#         "all triangulations should be symmetric, this is not"
#     )


def test_correctness_voronoi_clipping():
    noclip = _voronoi(lap_coords, clip=None, rook=True)
    extent = _voronoi(lap_coords, clip="extent", rook=True)
    alpha = _voronoi(lap_coords, clip="ashape", rook=True)

    g_noclip = Graph.from_arrays(*noclip)
    g_extent = Graph.from_arrays(*extent)
    g_alpha = Graph.from_arrays(*alpha)

    assert g_alpha < g_extent
    assert g_extent <= g_noclip

    extent_known = [
        np.array([0, 0, 0, 0, 1, 1, 1, 2, 2, 3, 3, 3, 4, 4]),
        np.array([1, 2, 3, 4, 0, 3, 4, 0, 3, 0, 1, 2, 0, 1]),
    ]
    alpha_known = [
        np.array([0, 0, 0, 1, 2, 2, 3, 3, 4, 4]),
        np.array([2, 3, 4, 4, 0, 3, 0, 2, 0, 1]),
    ]

    np.testing.assert_array_equal(
        g_extent.adjacency.index.get_level_values(0), extent_known[0]
    )
    np.testing.assert_array_equal(
        g_extent.adjacency.index.get_level_values(1), extent_known[1]
    )

    np.testing.assert_array_equal(
        g_alpha.adjacency.index.get_level_values(0), alpha_known[0]
    )
    np.testing.assert_array_equal(
        g_alpha.adjacency.index.get_level_values(1), alpha_known[1]
    )


def test_correctness_delaunay():
    head, tail, weight = _delaunay(cau_coords)

    known_head = np.array([0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 4, 4, 4])
    known_tail = np.array([1, 2, 4, 0, 2, 3, 0, 1, 3, 4, 1, 2, 4, 0, 2, 3])

    np.testing.assert_array_equal(known_head, head)
    np.testing.assert_array_equal(known_tail, tail)
    np.testing.assert_array_equal(np.ones(head.shape), weight)

    head, tail, weight = _delaunay(lap_coords)
    known_head = np.array([0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])
    known_tail = np.array([1, 2, 3, 4, 0, 3, 4, 0, 3, 4, 0, 1, 2, 0, 1, 2])

    np.testing.assert_array_equal(known_head, head)
    np.testing.assert_array_equal(known_tail, tail)
    np.testing.assert_array_equal(np.ones(head.shape), weight)


def test_correctness_voronoi():
    head, tail, weight = _voronoi(cau_coords)

    known_head = np.array([0, 0, 0, 1, 1, 2, 2, 2, 2, 3, 4, 4])
    known_tail = np.array([1, 2, 4, 0, 2, 0, 1, 3, 4, 2, 0, 2])

    np.testing.assert_array_equal(known_head, head)
    np.testing.assert_array_equal(known_tail, tail)
    np.testing.assert_array_equal(np.ones(head.shape), weight)

    head, tail, weight = _voronoi(lap_coords)
    known_head = np.array([0, 0, 0, 0, 1, 1, 1, 2, 2, 3, 3, 3, 4, 4])
    known_tail = np.array([1, 2, 3, 4, 0, 3, 4, 0, 3, 0, 1, 2, 0, 1])

    np.testing.assert_array_equal(known_head, head)
    np.testing.assert_array_equal(known_tail, tail)
    np.testing.assert_array_equal(np.ones(head.shape), weight)


def test_correctness_gabriel():
    head, tail, weight = _gabriel(cau_coords)

    known_head = np.array([0, 0, 1, 1, 2, 2, 2, 3, 4, 4])
    known_tail = np.array([1, 4, 0, 2, 1, 3, 4, 2, 0, 2])

    np.testing.assert_array_equal(known_head, head)
    np.testing.assert_array_equal(known_tail, tail)
    np.testing.assert_array_equal(np.ones(head.shape), weight)

    head, tail, weight = _gabriel(lap_coords)
    known_head = np.array([0, 0, 0, 1, 2, 2, 3, 3, 4, 4])
    known_tail = np.array([2, 3, 4, 4, 0, 3, 0, 2, 0, 1])

    np.testing.assert_array_equal(known_head, head)
    np.testing.assert_array_equal(known_tail, tail)
    np.testing.assert_array_equal(np.ones(head.shape), weight)


def test_correctness_relative_n():
    head, tail, weight = _relative_neighborhood(cau_coords)

    known_head = np.array([0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 4, 4, 4])
    known_tail = np.array([1, 2, 4, 0, 2, 3, 0, 1, 3, 4, 1, 2, 4, 0, 2, 3])

    np.testing.assert_array_equal(known_head, head)
    np.testing.assert_array_equal(known_tail, tail)
    np.testing.assert_array_equal(np.ones(head.shape), weight)

    head, tail, weight = _relative_neighborhood(lap_coords)
    known_head = np.array([0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4])
    known_tail = np.array([1, 2, 3, 4, 0, 4, 0, 3, 0, 2, 0, 1])

    np.testing.assert_array_equal(known_head, head)
    np.testing.assert_array_equal(known_tail, tail)
    np.testing.assert_array_equal(np.ones(head.shape), weight)


@parametrize_ids
def test_ids(ids):
    data = stores_unique.sample(frac=1)
    if ids is not None:
        data = data.set_index(ids)
    head, tail, weight = _delaunay(data)

    assert head.shape[0] == 3368
    assert tail.shape == head.shape
    assert weight.shape == head.shape
    np.testing.assert_array_equal(pd.unique(head), data.index)

    head, tail, weight = _voronoi(data)

    assert head.shape[0] == 3308
    assert tail.shape == head.shape
    assert weight.shape == head.shape
    np.testing.assert_array_equal(pd.unique(head), data.index)

    head, tail, weight = _gabriel(data)

    assert head.shape[0] == 2024
    assert tail.shape == head.shape
    assert weight.shape == head.shape
    np.testing.assert_array_equal(pd.unique(head), data.index)

    head, tail, weight = _relative_neighborhood(data)

    # assert head.shape[0] == 3346  # relativehood returns unstable results atm see #573
    assert tail.shape == head.shape
    assert weight.shape == head.shape
    np.testing.assert_array_equal(pd.unique(head), data.index)


def test_kernel():
    _, _, weight = _delaunay(cau_coords, kernel="gaussian")
    expected = np.array(
        [
            0.1305618,
            0.17359059,
            0.22312817,
            0.1305618,
            0.1339654,
            0.03259559,
            0.17359059,
            0.1339654,
            0.06948067,
            0.180294,
            0.03259559,
            0.06948067,
            0.02004257,
            0.22312817,
            0.180294,
            0.02004257,
        ]
    )

    np.testing.assert_array_almost_equal(expected, weight)


def test_coincident_raise_voronoi():
    with pytest.raises(ValueError, match="There are"):
        _voronoi(stores, clip=False)


def test_coincident_jitter_voronoi():
    cp_heads, cp_tails, cp_w = _voronoi(stores, clip=False, coincident="jitter")
    unique_heads, unique_tails, unique_w = _voronoi(stores_unique, clip=False)
    assert not np.array_equal(cp_heads, unique_heads)
    assert not np.array_equal(cp_tails, unique_tails)
    assert not np.array_equal(cp_w, unique_w)

    assert cp_heads.shape[0] == 3392
    assert unique_heads.shape[0] == 3368


def test_coincident_clique_voronoi():
    with pytest.raises(NotImplementedError, match="clique-based"):
        _voronoi(stores, clip=False, coincident="clique")
    # G_voronoi_cp_heads, G_voronoi_cp_tails, cp_w = _voronoi(
    #     stores, clip=False, coincident="clique"
    # )
    # G_voronoi_unique_heads, G_voronoi_unique_tails, unique_w = _voronoi(
    #     stores_unique, clip=False
    # )
    # assert not np.array_equal(G_voronoi_cp_heads, G_voronoi_unique_heads)
    # assert not np.array_equal(G_voronoi_cp_tails, G_voronoi_unique_tails)
    # assert not np.array_equal(cp_w, unique_w)


class TestCoincident:
    def setup_method(self):
        self.geom = [
            shapely.Point(0, 0),
            shapely.Point(1, 1),
            shapely.Point(2, 0),
            shapely.Point(3, 1),
            shapely.Point(0, 0),  # coincident point
            shapely.Point(0, 5),
        ]
        self.df_int = geopandas.GeoDataFrame(
            geometry=self.geom,
        )
        self.df_string = geopandas.GeoDataFrame(
            geometry=self.geom, index=["zero", "one", "two", "three", "four", "five"]
        )
        self.mapping = {0: "zero", 1: "one", 2: "two", 3: "three", 4: "four", 5: "five"}

    def test_delaunay_error(self):
        with pytest.raises(
            ValueError,
            match="There are 5 unique locations in the dataset, but 6 observations",
        ):
            _delaunay(self.df_int)

    def test_delaunay_jitter(self):
        heads, tails, weights = _delaunay(self.df_int, coincident="jitter", seed=0)

        exp_heads = np.array(
            [0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5]
        )
        exp_tails = np.array(
            [1, 2, 4, 0, 2, 3, 4, 5, 0, 1, 3, 1, 2, 5, 0, 1, 5, 1, 3, 4]
        )
        exp_w = np.ones(exp_heads.shape, dtype="int8")

        np.testing.assert_array_equal(heads, exp_heads)
        np.testing.assert_array_equal(tails, exp_tails)

        heads, tails, weights = _delaunay(self.df_string, coincident="jitter", seed=0)

        np.testing.assert_array_equal(heads, np.vectorize(self.mapping.get)(exp_heads))
        np.testing.assert_array_equal(tails, np.vectorize(self.mapping.get)(exp_tails))
        np.testing.assert_array_equal(weights, exp_w)

    @pytest.mark.xfail
    def test_delaunay_clique(self):
        # TODO: fix the implemntation to make this pass
        heads, tails, weights = _delaunay(self.df_int, coincident="clique", seed=0)

        exp_heads = np.array(
            [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5]
        )
        exp_tails = np.array(
            [1, 2, 4, 5, 0, 2, 3, 5, 0, 1, 3, 1, 2, 5, 0, 1, 2, 5, 0, 1, 3]
        )
        exp_w = np.ones(exp_heads.shape, dtype="int8")

        np.testing.assert_array_equal(heads, exp_heads)
        np.testing.assert_array_equal(tails, exp_tails)

        heads, tails, weights = _delaunay(self.df_string, coincident="clique", seed=0)

        np.testing.assert_array_equal(heads, np.vectorize(self.mapping.get)(exp_heads))
        np.testing.assert_array_equal(tails, np.vectorize(self.mapping.get)(exp_tails))
        np.testing.assert_array_equal(weights, exp_w)

    def test_gabriel_error(self):
        with pytest.raises(
            ValueError,
            match="There are 5 unique locations in the dataset, but 6 observations",
        ):
            _gabriel(self.df_int)

    def test_gabriel_jitter(self):
        heads, tails, weights = _gabriel(self.df_int, coincident="jitter", seed=0)

        exp_heads = np.array([0, 0, 1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 5, 5])
        exp_tails = np.array([2, 4, 2, 3, 4, 5, 0, 1, 3, 1, 2, 5, 0, 1, 1, 3])
        exp_w = np.ones(exp_heads.shape, dtype="int8")

        np.testing.assert_array_equal(heads, exp_heads)
        np.testing.assert_array_equal(tails, exp_tails)

        heads, tails, weights = _gabriel(self.df_string, coincident="jitter", seed=0)

        np.testing.assert_array_equal(heads, np.vectorize(self.mapping.get)(exp_heads))
        np.testing.assert_array_equal(tails, np.vectorize(self.mapping.get)(exp_tails))
        np.testing.assert_array_equal(weights, exp_w)

    @pytest.mark.xfail
    def test_gabriel_clique(self):
        # TODO: fix the implemntation to make this pass
        heads, tails, weights = _gabriel(self.df_int, coincident="clique", seed=0)

        exp_heads = np.array(
            [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5]
        )
        exp_tails = np.array(
            [1, 2, 4, 5, 0, 2, 3, 5, 0, 1, 3, 1, 2, 5, 0, 1, 2, 5, 0, 1, 3]
        )
        exp_w = np.ones(exp_heads.shape, dtype="int8")

        np.testing.assert_array_equal(heads, exp_heads)
        np.testing.assert_array_equal(tails, exp_tails)

        heads, tails, weights = _gabriel(self.df_string, coincident="clique", seed=0)

        np.testing.assert_array_equal(heads, np.vectorize(self.mapping.get)(exp_heads))
        np.testing.assert_array_equal(tails, np.vectorize(self.mapping.get)(exp_tails))
        np.testing.assert_array_equal(weights, exp_w)

    def test_relative_neighborhood_error(self):
        with pytest.raises(
            ValueError,
            match="There are 5 unique locations in the dataset, but 6 observations",
        ):
            _relative_neighborhood(self.df_int)

    def test_relative_neighborhood_jitter(self):
        heads, tails, weights = _relative_neighborhood(
            self.df_int, coincident="jitter", seed=0
        )

        exp_heads = np.array([0, 0, 0, 1, 1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5])
        exp_tails = np.array([1, 2, 4, 0, 3, 4, 5, 0, 3, 1, 2, 5, 0, 1, 5, 1, 3, 4])
        exp_w = np.ones(exp_heads.shape, dtype="int8")

        np.testing.assert_array_equal(heads, exp_heads)
        np.testing.assert_array_equal(tails, exp_tails)

        heads, tails, weights = _relative_neighborhood(
            self.df_string, coincident="jitter", seed=0
        )

        np.testing.assert_array_equal(heads, np.vectorize(self.mapping.get)(exp_heads))
        np.testing.assert_array_equal(tails, np.vectorize(self.mapping.get)(exp_tails))
        np.testing.assert_array_equal(weights, exp_w)

    @pytest.mark.xfail
    def test_relative_neighborhood_clique(self):
        # TODO: fix the implemntation to make this pass
        heads, tails, weights = _relative_neighborhood(
            self.df_int, coincident="clique", seed=0
        )

        exp_heads = np.array(
            [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5]
        )
        exp_tails = np.array(
            [1, 2, 4, 5, 0, 2, 3, 5, 0, 1, 3, 1, 2, 5, 0, 1, 2, 5, 0, 1, 3]
        )
        exp_w = np.ones(exp_heads.shape, dtype="int8")

        np.testing.assert_array_equal(heads, exp_heads)
        np.testing.assert_array_equal(tails, exp_tails)

        heads, tails, weights = _relative_neighborhood(
            self.df_string, coincident="clique", seed=0
        )

        np.testing.assert_array_equal(heads, np.vectorize(self.mapping.get)(exp_heads))
        np.testing.assert_array_equal(tails, np.vectorize(self.mapping.get)(exp_tails))
        np.testing.assert_array_equal(weights, exp_w)

    def test_voronoi_error(self):
        with pytest.raises(
            ValueError,
            match="There are 5 unique locations in the dataset, but 6 observations",
        ):
            _voronoi(self.df_int)

    def test_voronoi_jitter(self):
        heads, tails, weights = _voronoi(self.df_int, coincident="jitter", seed=0)

        exp_heads = np.array([0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 5, 5])
        exp_tails = np.array([1, 2, 4, 0, 2, 3, 4, 5, 0, 1, 3, 1, 2, 5, 0, 1, 1, 3])
        exp_w = np.ones(exp_heads.shape, dtype="int8")

        np.testing.assert_array_equal(heads, exp_heads)
        np.testing.assert_array_equal(tails, exp_tails)

        heads, tails, weights = _voronoi(self.df_string, coincident="jitter", seed=0)

        np.testing.assert_array_equal(heads, np.vectorize(self.mapping.get)(exp_heads))
        np.testing.assert_array_equal(tails, np.vectorize(self.mapping.get)(exp_tails))
        np.testing.assert_array_equal(weights, exp_w)

    @pytest.mark.xfail
    def test_voronoi_clique(self):
        # TODO: fix the implemntation to make this pass
        heads, tails, weights = _voronoi(self.df_int, coincident="clique", seed=0)

        exp_heads = np.array([0, 0, 1, 1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 5, 5])
        exp_tails = np.array([1, 4, 0, 2, 3, 5, 1, 3, 1, 2, 5, 0, 1, 1, 3])
        exp_w = np.ones(exp_heads.shape, dtype="int8")

        np.testing.assert_array_equal(heads, exp_heads)
        np.testing.assert_array_equal(tails, exp_tails)

        heads, tails, weights = _voronoi(self.df_string, coincident="clique", seed=0)

        np.testing.assert_array_equal(heads, np.vectorize(self.mapping.get)(exp_heads))
        np.testing.assert_array_equal(tails, np.vectorize(self.mapping.get)(exp_tails))
        np.testing.assert_array_equal(weights, exp_w)

    def test_wrong_resolver(self):
        with pytest.raises(
            ValueError,
            match="Recieved option coincident='nonsense'",
        ):
            _delaunay(self.df_int, coincident="nonsense")



import os
import platform

import pytest
import shapely

from .... import examples as pysal_examples
from ... import geotable as pdio
from ...fileio import FileIO as psopen

try:
    import sqlalchemy

    missing_sql = False
except ImportError:
    missing_sql = True


windows = platform.system() == "Windows"


@pytest.mark.skipif(windows, reason="Skipping Windows due to `PermissionError`.")
@pytest.mark.skipif(missing_sql, reason="Missing dependency: SQLAlchemy.")
class TestSqliteReader:
    def setup_method(self):
        path = pysal_examples.get_path("new_haven_merged.dbf")
        if path is None:
            pysal_examples.load_example("newHaven")
            path = pysal_examples.get_path("new_haven_merged.dbf")
        df = pdio.read_files(path)
        df["GEOMETRY"] = shapely.to_wkb(shapely.points(df["geometry"].values.tolist()))
        # This is a hack to not have to worry about a custom point type in the DB
        del df["geometry"]
        self.dbf = "iohandlers_test_db.db"
        engine = sqlalchemy.create_engine(f"sqlite:///{self.dbf}")
        self.conn = engine.connect()
        df.to_sql(
            "newhaven",
            self.conn,
            index=True,
            dtype={
                "date": sqlalchemy.types.UnicodeText,  # Should convert the df date into a true date object, just a hack again
                "dataset": sqlalchemy.types.UnicodeText,
                "street": sqlalchemy.types.UnicodeText,
                "intersection": sqlalchemy.types.UnicodeText,
                "time": sqlalchemy.types.UnicodeText,  # As above re: date
                "GEOMETRY": sqlalchemy.types.BLOB,
            },
        )  # This is converted to TEXT as lowest type common sqlite

    def test_deserialize(self):
        db = psopen(f"sqlite:///{self.dbf}")
        assert db.tables == ["newhaven"]

        gj = db._get_gjson("newhaven")
        assert gj["type"] == "FeatureCollection"

        self.conn.close()

        os.remove(self.dbf)



import os

import pytest

from ... import examples
from .. import user
from ..contiguity import Rook, Voronoi


class Testuser:
    def test_min_threshold_dist_from_shapefile(self):
        f = examples.get_path("columbus.shp")
        min_d = user.min_threshold_dist_from_shapefile(f)
        assert min_d == pytest.approx(0.61886415807685413)

    def test_build_lattice_shapefile(self):
        of = "lattice.shp"
        user.build_lattice_shapefile(20, 20, of)
        w = Rook.from_shapefile(of)
        assert w.n == 400
        os.remove("lattice.dbf")
        os.remove("lattice.shp")
        os.remove("lattice.shx")



import warnings

from . import (
    arcgis_dbf,
    arcgis_swm,
    arcgis_txt,
    csvWrapper,
    dat,
    gal,
    geobugs_txt,
    geoda_txt,
    gwt,
    mat,
    mtx,
    pyDbfIO,
    pyShpIO,
    stata_txt,
    wk1,
    wkt,
)

try:
    from . import db
except:  # noqa E722
    warnings.warn("SQLAlchemy not installed, database I/O disabled")  # noqa B028



"""
Quad Tree Structure Class for PySAL_core/cg/shapes.Ring, PySAL_core/cg/shapes.Polygon
This structure could speed up the determining of point in polygon process
"""
__author__ = "Hu Shao"

import math

from .shapes import Ring


def cwt(a, b, tolerance=1e-9):
    """compare_with_tolerance

    For the float value comparing, there are some situlation that two values
    are actually the same but been shown differently, e.g 1.230 and
    1.2300000000000001. Especially after some calculation. This function is
    used to compare two float value with some tolerance

    Parameters
    ----------
    a : float
        The first value
    b : float
        The second value
    tolerance : float
        Tolerance for the comparing

    Returns
    -------
    if_bigger_than  : int
        if a is bigger than b 1: a > b 0: a == b -1: a < b
    """
    tolerance = math.fabs(tolerance)
    if a - b > tolerance:
        return 1
    elif a - b < -tolerance:
        return -1
    else:
        return 0


class Cell:
    """Basic rectangle geometry used for dividing research
    area (polygon) into quadtree structure.

    Attributes
    ----------
    level : int
        Quadtree level this cell belongs to. Begins with 0
    min_x : float
        min x coordinate of this cell
    min_y : float
        min y coordinate of this cell
    length_x : float
        width of this cell
    length_y : float
        height of this cell
    arcs : list
        detected arc list which are within this cell
    status : str
        enum status of this cell, indicating this cell's spatial
        relationship with the research area
                  "in"      : this cell lies totally inside of the research area
                  "out"     : this cell lies totally outside of the research area
                  "maybe"   : this cell intersects with the research area's boundary
    children_l_b : Cell
        children of current cell, left-bottom
    children_l_t : Cell
        children of current cell, left-top
    children_r_b : Cell
        children of current cell, right-bottom
    children_t_t : Cell
        children of current cell, right-top
    """

    def __init__(self, level, min_x, min_y, length_x, length_y, arcs, status):
        """

        Parameters
        ----------
        level : int
            on which quadtree level this cell belongs to. Begins with 0
        min_x : float
            min x coordinate of this cell
        min_y : float
            min y coordinate of this cell
        length_x : float
            width of this cell
        length_y : float
            height of this cell
        arcs : list
            detected arc list which are within this cell
        status : str
            enum status of this cell, indicating this cell's spatial
            relationship with the research area
                      "in"      : this cell lies totally inside of the research area
                      "out"     : this cell lies totally outside of the research area
                      "maybe"   : this cell intersects with the research area's boundary
        """
        self.level = level
        self.min_x = min_x
        self.min_y = min_y
        self.length_x = length_x
        self.length_y = length_y
        self.arcs = arcs
        self.status = status
        self.zero_tolerance = 1e-9
        self._rings = None
        self.children_l_b = None
        self.children_l_t = None
        self.children_r_b = None
        self.children_r_t = None

    @property
    def rings(self):
        """the list of rings which are formed by the intersection
        of this cell and the arcs pass them

        Returns
        -------
        rings : list
        """
        if self._rings is None:
            if self.status == "in" or self.status == "out":
                self._rings = []
            else:
                self._rings = []
                extract_result = extract_segments_from_cell_with_arcs(
                    [self.min_x, self.min_y],
                    self.length_x,
                    self.length_y,
                    self.arcs,
                    self.zero_tolerance,
                )
                for point_list in extract_result[0]:
                    self._rings.append(Ring(point_list))

        return self._rings

    def split(self):
        """Equally split current cell into 4 sub cells. If this cell in needed
        to be splitted into four parts, add the result cells as children
        to current cell.
        """

        if self.status == "in" or self.status == "out":
            # no need to conduct the splitting
            return

        level = self.level + 1
        length_x = self.length_x / 2
        length_y = self.length_y / 2
        middle_x = self.min_x + length_x
        middle_y = self.min_y + length_y
        """
        Do the splitting work here.
        Some properties of the arc:
        - Point order of the arcs MUST be clockwise
        - The two end-points of each arc MUST lie on the borders of the cell
        - When a arc goes in a cell, it MUST goes out from the same one
        - The intersection points MUST be lying on the
            inner-boundaries which are used to divide the cell into 4 sub-cells
        - Use the intersection points to split the arcs into small ones
        - No need to store cell boundaries as arcs, store the
            intersection points, points' relative location from
        """
        if self.level == 0:
            if len(self.arcs) != 1:
                raise LookupError(
                    "Unexpected arc number! Only single ring "
                    "can be assigned to the root cell"
                )
                return
            # Do some initialize work, find one point lies on
            # the border of the rectangle(cell) and begin with this point
            arc = self.arcs[0]
            if (
                arc[0] == arc[len(arc) - 1]
            ):  # remove the duplicated points the the end of the arc
                arc = arc[0 : len(arc) - 1]
            min_x = arc[0][0]
            min_x_index = 0
            for index in range(0, len(arc)):
                if arc[index][0] < min_x:
                    min_x = arc[index][0]
                    min_x_index = index
            arc = arc[min_x_index : len(arc)] + arc[0 : min_x_index + 1]
            self.arcs[0] = arc

        # l: left, r: right, b: bottom, t: top
        cell_arcs_l_b = []
        cell_arcs_r_b = []
        cell_arcs_l_t = []
        cell_arcs_r_t = []
        for arc in self.arcs:
            temp_arc = []
            temp_arc_belonging = None
            for index in range(0, len(arc) - 1):
                x0 = arc[index][0]
                y0 = arc[index][1]
                x1 = arc[index + 1][0]
                y1 = arc[index + 1][1]
                if temp_arc_belonging is None:
                    """
                    In this section, determine which sub-cell does
                    the current temp_arc belong to. Although every
                    single arc must begin and end on the cell's outer
                    boundaries, when the split process begin, there might
                    be some sub-arcs begin at the split-line. So here we should
                    consider all possible situations See the image
                    cell_boundary_category_rule to better understand the process
                    """
                    if cwt(x0, self.min_x, self.zero_tolerance) == 0:  # left border
                        if cwt(y0, middle_y, self.zero_tolerance) == -1:  # position 1
                            temp_arc_belonging = "l_b"
                        elif cwt(y0, middle_y, self.zero_tolerance) == 1:  # position 2
                            temp_arc_belonging = "l_t"
                        else:  # just by chance at the middle point
                            if cwt(y1, y0, self.zero_tolerance) == -1:  # going down
                                temp_arc_belonging = "l_b"
                            elif cwt(y1, y0, self.zero_tolerance) == 1:  # going up
                                temp_arc_belonging = "l_t"
                            else:
                                # just by chance this segment lies on
                                # split_line_h, throw it
                                continue
                    elif (
                        cwt(x0, self.min_x + self.length_x, self.zero_tolerance) == 0
                    ):  # right border
                        if cwt(y0, middle_y, self.zero_tolerance) == -1:  # position 6
                            temp_arc_belonging = "r_b"
                        elif cwt(y0, middle_y, self.zero_tolerance) == 1:  # position 5
                            temp_arc_belonging = "r_t"
                        else:  # just by chance at the middle point
                            if cwt(y1, y0, self.zero_tolerance) == -1:  # going down
                                temp_arc_belonging = "r_b"
                            elif cwt(y1, y0, self.zero_tolerance) == 1:  # going up
                                temp_arc_belonging = "r_t"
                            else:
                                # just by chance this segment lies on
                                # split_line_h, throw it
                                continue
                    elif cwt(y0, self.min_y, self.zero_tolerance) == 0:  # bottom border
                        if cwt(x0, middle_x, self.zero_tolerance) == -1:  # position 8
                            temp_arc_belonging = "l_b"
                        elif cwt(x0, middle_x, self.zero_tolerance) == 1:  # position 7
                            temp_arc_belonging = "r_b"
                        else:  # just by chance at the middle point
                            if cwt(x1, x0, self.zero_tolerance) == -1:  # going left
                                temp_arc_belonging = "l_b"
                            elif cwt(x1, x0, self.zero_tolerance) == 1:  # going right
                                temp_arc_belonging = "r_b"
                            else:
                                # just by chance this segment lies on
                                # split_line_v, throw it
                                continue
                    elif (
                        cwt(y0, self.min_y + self.length_y, self.zero_tolerance) == 0
                    ):  # top border
                        if cwt(x0, middle_x, self.zero_tolerance) == -1:  # position 3
                            temp_arc_belonging = "l_t"
                        elif cwt(x0, middle_x, self.zero_tolerance) == 1:  # position 4
                            temp_arc_belonging = "r_t"
                        else:  # just by chance at the middle point
                            if cwt(x1, x0, self.zero_tolerance) == -1:  # going left
                                temp_arc_belonging = "l_t"
                            elif cwt(x1, x0, self.zero_tolerance) == 1:  # going right
                                temp_arc_belonging = "r_t"
                            else:
                                # just by chance this segment lies
                                # on split_line_v, throw it
                                continue
                    elif cwt(x0, middle_x, self.zero_tolerance) == 0:  # split_line_v
                        if cwt(y0, middle_y, self.zero_tolerance) == 1:  # position c
                            if cwt(x1, x0, self.zero_tolerance) == 1:
                                temp_arc_belonging = "r_t"
                            elif cwt(x1, x0, self.zero_tolerance) == -1:
                                temp_arc_belonging = "l_t"
                            else:  # x1==x0, just by chance on the split_line_v
                                continue
                        elif cwt(y0, middle_y, self.zero_tolerance) == -1:  # position d
                            if cwt(x1, x0, self.zero_tolerance) == 1:
                                temp_arc_belonging = "r_b"
                            elif cwt(x1, x0, self.zero_tolerance) == -1:
                                temp_arc_belonging = "l_b"
                            else:  # x1==x0, just by chance on the split_line_v
                                continue
                        else:
                            # in condition that p0 lies at the
                            # cross point of two split lines
                            if (
                                cwt(x1, x0, self.zero_tolerance) == 0
                                or cwt(y1, y0, self.zero_tolerance) == 0
                            ):  # on one of the split_line
                                continue
                            elif cwt(x1, x0, self.zero_tolerance) == 1:
                                if cwt(y1, y0, self.zero_tolerance) == 1:
                                    temp_arc_belonging = "r_t"
                                else:
                                    temp_arc_belonging = "r_b"
                            else:  # on condition that x1 < x0
                                if cwt(y1, y0, self.zero_tolerance) == 1:
                                    temp_arc_belonging = "l_t"
                                else:
                                    temp_arc_belonging = "l_b"
                    elif cwt(y0, middle_y, self.zero_tolerance) == 0:  # split_line_h
                        if cwt(x0, middle_x, self.zero_tolerance) == 1:  # position r
                            if cwt(y1, y0, self.zero_tolerance) == 1:
                                temp_arc_belonging = "r_t"
                            elif cwt(y1, y0, self.zero_tolerance) == -1:
                                temp_arc_belonging = "r_b"
                            else:  # y1==y0, just by chance on the split_line_h
                                continue
                        else:  # on condition that x0 < middle_x, position a
                            if cwt(y1, y0, self.zero_tolerance) == 1:
                                temp_arc_belonging = "l_t"
                            elif cwt(y1, y0, self.zero_tolerance) == -1:
                                temp_arc_belonging = "l_b"
                            else:  # y1==y0, just by chance on the split_line_h
                                continue
                if temp_arc_belonging is None:
                    raise Exception("Error on cell split!!!")

                # At this point, the belonging sub-cell
                # of current segment is already known.
                # Let's begin the splitting!
                """
                Firstly determine if the segment totally lies on the split_lines.
                p1 (x1, y1) could lie on the split_lines. This situation is not
                the same with previous ones which "both points lie on the same
                split_line". In previous situation, the p1 is the begin point of a
                sub arc, we can just throw that segment if it totally lie on
                split_line. However, in current situation, the segment is in the
                middle of the sub-arc. So if the segment is detected totally lie
                on the split_line, we should split the sub_arc here.
                """
                if (
                    cwt(x0, x1, self.zero_tolerance)
                    == cwt(x0, middle_x, self.zero_tolerance)
                    == 0
                    or cwt(y0, y1, self.zero_tolerance)
                    == cwt(y0, middle_y, self.zero_tolerance)
                    == 0
                ):  # split the arc here, throw current segment
                    if len(temp_arc) != 0:
                        if temp_arc_belonging == "l_b":
                            cell_arcs_l_b.append(temp_arc)
                        elif temp_arc_belonging == "l_t":
                            cell_arcs_l_t.append(temp_arc)
                        elif temp_arc_belonging == "r_b":
                            cell_arcs_r_b.append(temp_arc)
                        elif temp_arc_belonging == "r_t":
                            cell_arcs_r_t.append(temp_arc)
                    temp_arc = []
                    temp_arc_belonging = None
                    continue

                intersect_point_h = None
                intersect_point_v = None
                # Check if the segment intersects with split_line_h
                if (
                    cwt(y0, middle_y, self.zero_tolerance) == -1
                    and cwt(middle_y, y1, self.zero_tolerance) <= 0
                ) or (
                    cwt(y0, middle_y, self.zero_tolerance) == 1
                    and cwt(middle_y, y1, self.zero_tolerance) >= 0
                ):
                    if (
                        cwt(x0, x1, self.zero_tolerance) == 0
                    ):  # the segments is vertical
                        intersect_point_h = [x0, middle_y]
                    else:
                        a = (y1 - y0) / (x1 - x0)
                        b = y0 - a * x0
                        x_new = (middle_y - b) / a
                        intersect_point_h = [x_new, middle_y]
                # Check if the segment intersects with split_line_v
                if (
                    cwt(x0, middle_x, self.zero_tolerance) == -1
                    and cwt(middle_x, x1, self.zero_tolerance) <= 0
                ) or (
                    cwt(x0, middle_x, self.zero_tolerance) == 1
                    and cwt(middle_x, x1, self.zero_tolerance) >= 0
                ):
                    if (
                        cwt(y0, y1, self.zero_tolerance) == 0
                    ):  # the segments is horizontal
                        intersect_point_v = [middle_x, y0]
                    else:
                        a = (y1 - y0) / (x1 - x0)
                        b = y0 - a * x0
                        y_new = a * middle_x + b
                        intersect_point_v = [middle_x, y_new]
                # check if the intersect point(s) exist
                intersect_point = None
                intersect_point_mark = None
                if (intersect_point_h is not None) and (intersect_point_v is not None):
                    # In this situation, the current segment
                    # cannot be vertical nor horizontal.
                    # Find the closer intersection point to p0
                    if math.fabs(intersect_point_h[0] - x0) < math.fabs(
                        intersect_point_v[0] - x0
                    ):
                        intersect_point = intersect_point_h
                        intersect_point_mark = "h"
                    else:
                        intersect_point = intersect_point_v
                        intersect_point_mark = "v"
                elif intersect_point_h is not None:
                    intersect_point = intersect_point_h
                    intersect_point_mark = "h"
                elif intersect_point_v is not None:
                    intersect_point = intersect_point_v
                    intersect_point_mark = "v"

                if intersect_point is not None:  # split the arc here
                    if len(temp_arc) == 0:
                        temp_arc.append([x0, y0])
                    temp_arc.append(intersect_point)
                    if temp_arc_belonging == "l_b":
                        cell_arcs_l_b.append(temp_arc)
                    elif temp_arc_belonging == "l_t":
                        cell_arcs_l_t.append(temp_arc)
                    elif temp_arc_belonging == "r_b":
                        cell_arcs_r_b.append(temp_arc)
                    elif temp_arc_belonging == "r_t":
                        cell_arcs_r_t.append(temp_arc)

                    if temp_arc_belonging == "l_b":
                        if intersect_point_mark == "h":
                            if (intersect_point_h is not None) and (
                                intersect_point_v is not None
                            ):
                                # under the situation that a single segment
                                # intersects with both split lines,
                                # here need carefully process
                                small_arc = [intersect_point_h, intersect_point_v]
                                if (
                                    cwt(
                                        intersect_point_h[0],
                                        intersect_point_v[0],
                                        self.zero_tolerance,
                                    )
                                    != 0
                                ):
                                    # deal with the situation that the segment
                                    # just goes through center point
                                    cell_arcs_l_t.append(small_arc)
                                temp_arc = [intersect_point_v, [x1, y1]]
                                temp_arc_belonging = "r_t"
                            else:
                                temp_arc = [intersect_point, [x1, y1]]
                                temp_arc_belonging = "l_t"
                        else:
                            if (intersect_point_h is not None) and (
                                intersect_point_v is not None
                            ):
                                # under the situation that a single segment
                                # intersects with both split lines,
                                # here need carefully process
                                small_arc = [intersect_point_v, intersect_point_h]
                                if (
                                    cwt(
                                        intersect_point_h[0],
                                        intersect_point_v[0],
                                        self.zero_tolerance,
                                    )
                                    != 0
                                ):
                                    # deal with the situation that
                                    # the segment just goes through center point
                                    cell_arcs_r_b.append(small_arc)
                                temp_arc = [intersect_point_h, [x1, y1]]
                                temp_arc_belonging = "r_t"
                            else:
                                temp_arc = [intersect_point, [x1, y1]]
                                temp_arc_belonging = "r_b"

                    elif temp_arc_belonging == "l_t":
                        if intersect_point_mark == "h":
                            if (intersect_point_h is not None) and (
                                intersect_point_v is not None
                            ):
                                # under the situation that a single segment
                                # intersects with both split lines,
                                # here need carefully process
                                small_arc = [intersect_point_h, intersect_point_v]
                                if (
                                    cwt(
                                        intersect_point_h[0],
                                        intersect_point_v[0],
                                        self.zero_tolerance,
                                    )
                                    != 0
                                ):
                                    # deal with the situation that the segment
                                    # just goes through center point
                                    cell_arcs_l_b.append(small_arc)
                                temp_arc = [intersect_point_v, [x1, y1]]
                                temp_arc_belonging = "r_b"
                            else:
                                temp_arc = [intersect_point, [x1, y1]]
                                temp_arc_belonging = "l_b"
                        else:
                            if (intersect_point_h is not None) and (
                                intersect_point_v is not None
                            ):
                                # under the situation that a single segment
                                # intersects with both split lines,
                                # here need carefully process
                                small_arc = [intersect_point_v, intersect_point_h]
                                if (
                                    cwt(
                                        intersect_point_h[0],
                                        intersect_point_v[0],
                                        self.zero_tolerance,
                                    )
                                    != 0
                                ):
                                    # deal with the situation that
                                    # the segment just goes through center point
                                    cell_arcs_r_t.append(small_arc)
                                temp_arc = [intersect_point_h, [x1, y1]]
                                temp_arc_belonging = "r_b"
                            else:
                                temp_arc = [intersect_point, [x1, y1]]
                                temp_arc_belonging = "r_t"

                    elif temp_arc_belonging == "r_b":
                        if intersect_point_mark == "h":
                            if (intersect_point_h is not None) and (
                                intersect_point_v is not None
                            ):
                                # under the situation that a single segment
                                # intersects with both split lines,
                                # here need carefully process
                                small_arc = [intersect_point_h, intersect_point_v]
                                if (
                                    cwt(
                                        intersect_point_h[0],
                                        intersect_point_v[0],
                                        self.zero_tolerance,
                                    )
                                    != 0
                                ):
                                    # deal with the situation that
                                    # the segment just goes through center point
                                    cell_arcs_r_t.append(small_arc)
                                temp_arc = [intersect_point_v, [x1, y1]]
                                temp_arc_belonging = "l_t"
                            else:
                                temp_arc = [intersect_point, [x1, y1]]
                                temp_arc_belonging = "r_t"
                        else:
                            if (intersect_point_h is not None) and (
                                intersect_point_v is not None
                            ):
                                # under the situation that a single segment
                                # intersects with both split lines,
                                # here need carefully process
                                small_arc = [intersect_point_v, intersect_point_h]
                                if (
                                    cwt(
                                        intersect_point_h[0],
                                        intersect_point_v[0],
                                        self.zero_tolerance,
                                    )
                                    != 0
                                ):
                                    # deal with the situation that
                                    # the segment just goes through center point
                                    cell_arcs_l_b.append(small_arc)
                                temp_arc = [intersect_point_h, [x1, y1]]
                                temp_arc_belonging = "l_t"
                            else:
                                temp_arc = [intersect_point, [x1, y1]]
                                temp_arc_belonging = "l_b"

                    elif temp_arc_belonging == "r_t":
                        if intersect_point_mark == "h":
                            if (intersect_point_h is not None) and (
                                intersect_point_v is not None
                            ):
                                # under the situation that a single segment
                                # intersects with both split lines,
                                # here need carefully process
                                small_arc = [intersect_point_h, intersect_point_v]
                                if (
                                    cwt(
                                        intersect_point_h[0],
                                        intersect_point_v[0],
                                        self.zero_tolerance,
                                    )
                                    != 0
                                ):
                                    # deal with the situation that
                                    # the segment just goes through center point
                                    cell_arcs_r_b.append(small_arc)
                                temp_arc = [intersect_point_v, [x1, y1]]
                                temp_arc_belonging = "l_b"
                            else:
                                temp_arc = [intersect_point, [x1, y1]]
                                temp_arc_belonging = "r_b"
                        else:
                            if (intersect_point_h is not None) and (
                                intersect_point_v is not None
                            ):
                                # under the situation that a single segment
                                # intersects with both split lines,
                                # here need carefully process
                                small_arc = [intersect_point_v, intersect_point_h]
                                if (
                                    cwt(
                                        intersect_point_h[0],
                                        intersect_point_v[0],
                                        self.zero_tolerance,
                                    )
                                    != 0
                                ):
                                    # deal with the situation that
                                    # the segment just goes through center point
                                    cell_arcs_l_t.append(small_arc)
                                temp_arc = [intersect_point_h, [x1, y1]]
                                temp_arc_belonging = "l_b"
                            else:
                                temp_arc = [intersect_point, [x1, y1]]
                                temp_arc_belonging = "l_t"
                    if (
                        cwt(temp_arc[0][0], temp_arc[1][0], self.zero_tolerance) == 0
                        and cwt(temp_arc[0][1], temp_arc[1][1], self.zero_tolerance)
                        == 0
                    ):
                        # to deal with the situation that p1
                        # just lied on one of the split-lines
                        temp_arc = []
                        temp_arc_belonging = None
                else:  # simply append the point to current arc
                    if len(temp_arc) == 0:
                        temp_arc.append([x0, y0])
                    temp_arc.append([x1, y1])
            # Allocate the last left arc to a sub-cell
            if len(temp_arc) > 0:
                if temp_arc_belonging == "l_b":
                    cell_arcs_l_b.append(temp_arc)
                elif temp_arc_belonging == "l_t":
                    cell_arcs_l_t.append(temp_arc)
                elif temp_arc_belonging == "r_b":
                    cell_arcs_r_b.append(temp_arc)
                elif temp_arc_belonging == "r_t":
                    cell_arcs_r_t.append(temp_arc)
        status_l_b = "maybe"
        status_l_t = "maybe"
        status_r_b = "maybe"
        status_r_t = "maybe"
        """
        At this point, all the arcs in this cell have been split into
        sub-arcs and allocated to 4 sub-cells. So, we can try to create
        the cells on left-bottom, right-bottom, left-top and right-top.
        Before doing that, we need to determine the status of each sub-cell,
        especially those who are totally within or out of the study area.
        These two kind of sub-cell have the same property: they don't have
        arc allocated. So, If here exists cell(s) who don't have arcs
        allocated, we need to begin the check
        """
        if (
            len(cell_arcs_l_b)
            * len(cell_arcs_l_t)
            * len(cell_arcs_r_b)
            * len(cell_arcs_r_t)
            == 0
        ):
            extract_result = extract_segments_from_cell_with_arcs(
                [self.min_x, self.min_y],
                self.length_x,
                self.length_y,
                self.arcs,
                self.zero_tolerance,
            )
            construct_rings = []
            for arc in extract_result[0]:
                construct_rings.append(Ring(arc))
            # determine the totally within and out-of sub cells
            if len(cell_arcs_l_b) == 0:
                center = [self.min_x + length_x / 2, self.min_y + length_y / 2]
                is_in = False
                for ring in construct_rings:
                    if ring.contains_point(center):
                        is_in = True
                status_l_b = "in" if is_in else "out"
            if len(cell_arcs_l_t) == 0:
                center = [self.min_x + length_x / 2, middle_y + length_y / 2]
                is_in = False
                for ring in construct_rings:
                    if ring.contains_point(center):
                        is_in = True
                status_l_t = "in" if is_in else "out"
            if len(cell_arcs_r_b) == 0:
                center = [middle_x + length_x / 2, self.min_y + length_y / 2]
                is_in = False
                for ring in construct_rings:
                    if ring.contains_point(center):
                        is_in = True
                status_r_b = "in" if is_in else "out"
            if len(cell_arcs_r_t) == 0:
                center = [middle_x + length_x / 2, middle_y + length_y / 2]
                is_in = False
                for ring in construct_rings:
                    if ring.contains_point(center):
                        is_in = True
                status_r_t = "in" if is_in else "out"

        cells_l_b = Cell(
            level, self.min_x, self.min_y, length_x, length_y, cell_arcs_l_b, status_l_b
        )
        cells_l_t = Cell(
            level, self.min_x, middle_y, length_x, length_y, cell_arcs_l_t, status_l_t
        )
        cells_r_b = Cell(
            level, middle_x, self.min_y, length_x, length_y, cell_arcs_r_b, status_r_b
        )
        cells_r_t = Cell(
            level, middle_x, middle_y, length_x, length_y, cell_arcs_r_t, status_r_t
        )
        self.children_l_b = cells_l_b
        self.children_l_t = cells_l_t
        self.children_r_b = cells_r_b
        self.children_r_t = cells_r_t

    #     ===================

    def contains_point(self, point):
        """Decide if this cell (rectangle) contains a given point

        Parameters
        ----------
        point : list
            Point structure, like [x, y]

        Returns
        -------
        if_contains : bool
        """
        if self.status == "out":
            return False
        if (
            point[0] < self.min_x
            or point[0] > self.min_x + self.length_x
            or point[1] < self.min_y
            or point[1] > self.min_y + self.length_y
        ):
            return False
        if self.status == "in":
            return True
        else:
            is_in = False
            for ring in self.rings:
                if ring.contains_point(point):
                    is_in = True
            return is_in


def extract_connecting_borders_between_points(
    cell_min_point, cell_length_x, cell_length_y, point_begin, point_end, zero_tolerance
):
    """There is an rectangle and two points on the border, this function is
    used to extract the borders connecting these two points. The segments must
    be clockwise

    Parameters
    ----------
    cell_min_point : list
        the bottom-left point of the cell, like [x0, y0]
    cell_length_x : float
        width of the cell
    cell_length_y : float
        height of the cell
    point_begin  : list
        the first point on the cell's border. like [xa, ya]
    point_end : list
        the second point on the cell's border. like [xb, yb]
    result_type : str
        MUST be one of ["segments", "border_ids"]. Indicts which kind of result
        will return. "segments": return the segments list which connecting
        these two points "border_ids" return a list of ids of the orders of the
        cell connecting these two points
    zero_tolerance : float
        value of zero_tolerance for determining if two float values are equal

    Returns
    -------
    segments_and_ids : tuple
        like (segments, involved_border_ids) 1. list of points, including the
        start and end points 2. list of border ids being involved in the
        segments, not necessary to be in the original order
    """
    if point_begin == point_end:
        return ([], [])
    # Determine which borders do the point_begin and point_end belong
    border_id_p_begin = -1
    border_id_p_end = -1
    if cwt(point_begin[0], cell_min_point[0], zero_tolerance) == 0:
        border_id_p_begin = 0
    elif cwt(point_begin[1], cell_min_point[1] + cell_length_y, zero_tolerance) == 0:
        border_id_p_begin = 1
    elif cwt(point_begin[0], cell_min_point[0] + cell_length_x, zero_tolerance) == 0:
        border_id_p_begin = 2
    elif cwt(point_begin[1], cell_min_point[1], zero_tolerance) == 0:
        border_id_p_begin = 3

    if cwt(point_end[0], cell_min_point[0], zero_tolerance) == 0:
        border_id_p_end = 0
    elif cwt(point_end[1], cell_min_point[1] + cell_length_y, zero_tolerance) == 0:
        border_id_p_end = 1
    elif cwt(point_end[0], cell_min_point[0] + cell_length_x, zero_tolerance) == 0:
        border_id_p_end = 2
    elif cwt(point_end[1], cell_min_point[1], zero_tolerance) == 0:
        border_id_p_end = 3

    if border_id_p_begin == -1 or border_id_p_end == -1:
        print(
            (
                cell_min_point,
                cell_min_point[0] + cell_length_x,
                cell_min_point[1] + cell_length_y,
                point_begin,
                point_end,
                cell_length_x,
                cell_length_y,
            )
        )
        raise Exception("Error! begin/end point doesn't lie on the cell border!!!")

    # Now, move forward from point_begin to point_end
    segments = [point_begin]
    involved_border_ids = [border_id_p_begin]
    border_id_p_search = border_id_p_begin
    if (
        border_id_p_search == border_id_p_end
    ):  # first check if they lie on the same border at the beginning
        if border_id_p_search == 0:
            if cwt(point_begin[1], point_end[1], zero_tolerance) == -1:
                segments.append(point_end)
                return (segments, involved_border_ids)
            else:
                segments.append([cell_min_point[0], cell_min_point[1] + cell_length_y])
                border_id_p_search = (border_id_p_search + 1) % 4
        elif border_id_p_search == 1:
            if cwt(point_begin[0], point_end[0], zero_tolerance) == -1:
                segments.append(point_end)
                return (segments, involved_border_ids)
            else:
                segments.append(
                    [
                        cell_min_point[0] + cell_length_x,
                        cell_min_point[1] + cell_length_y,
                    ]
                )
                border_id_p_search = (border_id_p_search + 1) % 4
        elif border_id_p_search == 2:
            if cwt(point_begin[1], point_end[1], zero_tolerance) == 1:
                segments.append(point_end)
                return (segments, involved_border_ids)
            else:
                segments.append([cell_min_point[0] + cell_length_x, cell_min_point[1]])
                border_id_p_search = (border_id_p_search + 1) % 4
        elif border_id_p_search == 3:
            if cwt(point_begin[0], point_end[0], zero_tolerance) == 1:
                segments.append(point_end)
                return (segments, involved_border_ids)
            else:
                segments.append([cell_min_point[0], cell_min_point[1]])
                border_id_p_search = (border_id_p_search + 1) % 4
    while True:
        involved_border_ids.append(border_id_p_search)
        if border_id_p_search != border_id_p_end:  # add a whole border
            if border_id_p_search == 0:
                segments.append([cell_min_point[0], cell_min_point[1] + cell_length_y])
            elif border_id_p_search == 1:
                segments.append(
                    [
                        cell_min_point[0] + cell_length_x,
                        cell_min_point[1] + cell_length_y,
                    ]
                )
            elif border_id_p_search == 2:
                segments.append([cell_min_point[0] + cell_length_x, cell_min_point[1]])
            elif border_id_p_search == 3:
                segments.append([cell_min_point[0], cell_min_point[1]])
            border_id_p_search = (border_id_p_search + 1) % 4
        else:  # add the border segment according to the enc point
            segments.append(point_end)
            return (segments, list(set(involved_border_ids)))


def get_relative_location_on_cell_border(
    cell_min_point, cell_length_x, cell_length_y, point, zero_tolerance
):
    """When a point is on the border of a cell, this function can be used to
    calculate the relative location of the point to cell's left-bottom corner.

    Parameters
    ----------
    cell_min_point : list
        the bottom-left point of the cell, like [x0, y0]
    cell_length_x : float
        width of the cell
    cell_length_y : float
        height of the cell
    point : list
         the point on the cell's border. like [x, y]
    zero_tolerance : float
        value of zero_tolerance for determining if two float values are equal

    Returns
    -------
    distance : float
        range from 0 to 4
    """
    border_id_p = -1
    if cwt(point[0], cell_min_point[0], zero_tolerance) == 0:
        border_id_p = 0
        border_id_p += (point[1] - cell_min_point[1]) / cell_length_y
    elif cwt(point[1], cell_min_point[1] + cell_length_y, zero_tolerance) == 0:
        border_id_p = 1
        border_id_p += (point[0] - cell_min_point[0]) / cell_length_x
    elif cwt(point[0], cell_min_point[0] + cell_length_x, zero_tolerance) == 0:
        border_id_p = 2
        border_id_p += 1 - (point[1] - cell_min_point[1]) / cell_length_y
    elif cwt(point[1], cell_min_point[1], zero_tolerance) == 0:
        border_id_p = 3
        border_id_p += 1 - (point[0] - cell_min_point[0]) / cell_length_x
    return border_id_p


def extract_segments_from_cell_with_arcs(
    cell_min_point, cell_length_x, cell_length_y, arcs, zero_tolerance
):
    """At the end of study area quadtree dividing, there will be some node
    cells intersect with arcs. The arcs are segments of original study border
    and the begin and end points of the arcs MUST lie on node cell border. This
    function can intersect the node cell and

    Parameters
    ----------
    cell_min_point : array
        the bottom-left point of the cell, like [x0, y0]
    cell_length_x : float
        width of the cell
    cell_length_y : float
        height of the cell
    arcs : array
        array of point lists
    zero_tolerance : float
        value of zero_tolerance for determining if two float values are equal

    Returns
    -------
    rings_and_border_ids : tuple
        like (rings, involved_border_ids) 1. the list of rings extracted, each
        ring contains a sequence of points - the begin and end points are the
        same. Note that there might be multiple rings extracted in a cell. 2.
        the ids of borders of the cell who are involved in the ring. Duplicated
        ids are removed and they may not be in the original order
    """
    arc_begin_points = []  # beginning points of each arc
    arc_begin_points_location = []  # location of beginning points of each arc
    arc_end_points = []  # ending points of each arc
    arc_end_points_location = []  # location of ending points of each arc
    for single_arc in arcs:
        arc_begin_points.append(single_arc[0])
        arc_end_points.append(single_arc[len(single_arc) - 1])
        arc_begin_points_location.append(
            get_relative_location_on_cell_border(
                cell_min_point,
                cell_length_x,
                cell_length_y,
                single_arc[0],
                zero_tolerance,
            )
        )
        arc_end_points_location.append(
            get_relative_location_on_cell_border(
                cell_min_point,
                cell_length_x,
                cell_length_y,
                single_arc[len(single_arc) - 1],
                zero_tolerance,
            )
        )

    rings = []
    involved_border_ids = []
    used_arc_ids = []
    # every time find an unused arc with minimum begin-point-location,
    # and begin the track form here
    new_ring = []  # the point list
    new_ring_end_point = None
    new_ring_end_point_location = -1
    selected_arc_ids = []
    while len(used_arc_ids) < len(arcs):
        if len(selected_arc_ids) == 0:
            # init the process of constructing a new ring
            # find the unused arc with min begin point location
            arc_id_with_min_begin_location = -1
            for i in range(0, len(arcs)):
                if i in used_arc_ids:
                    continue
                if (
                    arc_id_with_min_begin_location == -1
                    or cwt(
                        arc_begin_points_location[arc_id_with_min_begin_location],
                        arc_begin_points_location[i],
                        zero_tolerance,
                    )
                    == 1
                ):
                    arc_id_with_min_begin_location = i
            new_ring_end_point = arc_end_points[arc_id_with_min_begin_location]
            new_ring_end_point_location = arc_end_points_location[
                arc_id_with_min_begin_location
            ]
            selected_arc_ids.append(arc_id_with_min_begin_location)
            for point in arcs[arc_id_with_min_begin_location]:
                new_ring.append(point)
        else:
            # there is already a selected arc, find the next
            # available arc(maybe itself) and add the borders between
            # these two arcs
            arc_id_with_relatively_min_begin_location = -1
            for i in range(0, len(arcs)):
                if i in used_arc_ids:
                    continue
                if arc_id_with_relatively_min_begin_location == -1:
                    arc_id_with_relatively_min_begin_location = i
                else:
                    distance_to_end_point_min = (
                        arc_begin_points_location[
                            arc_id_with_relatively_min_begin_location
                        ]
                        - new_ring_end_point_location
                    )
                    if distance_to_end_point_min < 0:
                        distance_to_end_point_min += 4
                    distance_to_end_point_now = (
                        arc_begin_points_location[i] - new_ring_end_point_location
                    )
                    if distance_to_end_point_now < 0:
                        distance_to_end_point_now += 4
                    if (
                        cwt(
                            distance_to_end_point_min,
                            distance_to_end_point_now,
                            zero_tolerance,
                        )
                        == 1
                    ):
                        arc_id_with_relatively_min_begin_location = i
            extract_result = extract_connecting_borders_between_points(
                cell_min_point,
                cell_length_x,
                cell_length_y,
                new_ring_end_point,
                arc_begin_points[arc_id_with_relatively_min_begin_location],
                zero_tolerance,
            )
            point_list = extract_result[0]
            border_id_list = extract_result[1]
            for i in range(0, len(point_list)):
                if i == 0 and point_list[i] == new_ring[len(new_ring) - 1]:
                    continue
                new_ring.append(point_list[i])
            for border_id in border_id_list:
                if border_id not in involved_border_ids:
                    involved_border_ids.append(border_id)
            new_ring_end_point = arc_end_points[
                arc_id_with_relatively_min_begin_location
            ]
            new_ring_end_point_location = arc_end_points_location[
                arc_id_with_relatively_min_begin_location
            ]
            if arc_id_with_relatively_min_begin_location not in selected_arc_ids:
                #  find a new arc, add the point sequence
                # in this arc to the ring, and continue the searching further
                selected_arc_ids.append(arc_id_with_relatively_min_begin_location)
                single_arc = arcs[arc_id_with_relatively_min_begin_location]
                for i in range(0, len(single_arc)):
                    if i == 0 and single_arc[i] == new_ring[len(new_ring) - 1]:
                        continue
                    new_ring.append(single_arc[i])
            else:
                # the newly found arc is exactly the beginning one,
                # a whole closed ring is formed. stop here
                rings.append(new_ring)
                new_ring = []
                new_ring_end_point = None
                new_ring_end_point_location = -1
                for arc_id in selected_arc_ids:
                    used_arc_ids.append(arc_id)
                selected_arc_ids = []
    if len(new_ring) > 0:
        raise Exception("Error in extract_segments_from_cell_with_arcs!!!")
    return (rings, involved_border_ids)


class QuadTreeStructureSingleRing:
    """This class is the main manager of cells.

    By giving a study area. This class can construct a cell list depicting the
    study area. When given a new point. This class could rapidly determine
    whether the point lies in the study area

    Attributes
    __________
    root_cell : Cell
        The Cell structure for storing the quad-tree for this ring
    """

    def __init__(self, ring):
        """
        Parameters
        ----------
        ring : Ring
            the point list of study area. But in the class of Ring in PySAL
            Example: Ring([[0.0, 0.0], [3.0, 2.0], [5.0, 1.0]])
        """
        self.ring = ring
        self.root_cell = Cell(
            0,
            ring.bounding_box.left,
            ring.bounding_box.lower,
            ring.bounding_box.width,
            ring.bounding_box.height,
            [ring.vertices],
            "maybe",
        )

        # here build the quad tree structure
        # The criterion of stopping splitting the tree:
        #    1. The status is "in" or "out"
        #    2. The level >= 5 and the the number of current
        #       cell only contains one ring and all segments
        #       of the ring is no more than 4
        #    3. The level >= 8
        cells_for_processing = [self.root_cell]
        total_cell_count = 1
        for _i in range(0, 8):  # 10
            result_cell_list = []
            while len(cells_for_processing) > 0:
                cell = cells_for_processing.pop()
                cell.split()
                total_cell_count += 4
                children_cells = [
                    cell.children_l_b,
                    cell.children_l_t,
                    cell.children_r_b,
                    cell.children_r_t,
                ]
                for child in children_cells:
                    if child.status == "out" or child.status == "in":
                        continue
                    if child.level >= 5 and (  # 6
                        len(child.rings) == 1 and child.rings[0].len <= 5
                    ):
                        continue
                    result_cell_list.append(child)
            cells_for_processing = result_cell_list

    def contains_point(self, point):
        """Quickly determine if the study area contains a point

        Parameters
        ----------
        point : list
            the point structure, like [x, y]

        Returns
        -------
        if_contains : bool
        """

        # bbox check
        if (
            point[0] < self.min_x
            or point[0] > self.min_x + self.region_width
            or point[1] < self.min_y
            or point[1] > self.min_y + self.region_height
        ):
            return False
        # find the leaf cell for checking
        cell_to_check = self.root_cell
        while True:
            if cell_to_check.children_l_b is None:
                break
            middle_x = cell_to_check.min_x + cell_to_check.length_x / 2
            middle_y = cell_to_check.min_y + cell_to_check.length_y / 2
            if point[0] <= middle_x and point[1] <= middle_y:
                cell_to_check = cell_to_check.children_l_b
            elif point[0] <= middle_x and point[1] > middle_y:
                cell_to_check = cell_to_check.children_l_t
            elif point[0] > middle_x and point[1] <= middle_y:
                cell_to_check = cell_to_check.children_r_b
            else:
                cell_to_check = cell_to_check.children_r_t

        return cell_to_check.contains_point(point)

    @property
    def region_width(self):
        return self.ring.bounding_box.width

    @property
    def region_height(self):
        return self.ring.bounding_box.height

    @property
    def min_x(self):
        return self.ring.bounding_box.left

    @property
    def min_y(self):
        return self.ring.bounding_box.lower



import os
import tempfile

import pytest

from .... import examples as pysal_examples
from ...fileio import FileIO as psopen
from ..stata_txt import StataTextIO


class Testtest_StataTextIO:
    def setup_method(self):
        self.test_file_sparse = test_file_sparse = pysal_examples.get_path(
            "stata_sparse.txt"
        )
        self.test_file_full = test_file_full = pysal_examples.get_path("stata_full.txt")
        self.obj_sparse = StataTextIO(test_file_sparse, "r")
        self.obj_full = StataTextIO(test_file_full, "r")

    def test_close(self):
        for obj in [self.obj_sparse, self.obj_full]:
            f = obj
            f.close()
            pytest.raises(ValueError, f.read)

    def test_read(self):
        w_sparse = self.obj_sparse.read()
        assert w_sparse.n == 56
        assert w_sparse.mean_neighbors == 4.0
        assert [1.0, 1.0, 1.0, 1.0, 1.0] == list(w_sparse[1].values())

        w_full = self.obj_full.read()
        assert w_full.n == 56
        assert w_full.mean_neighbors == 4.0
        assert [0.125, 0.125, 0.125, 0.125, 0.125] == list(w_full[1].values())

    def test_seek(self):
        self.test_read()
        pytest.raises(StopIteration, self.obj_sparse.read)
        pytest.raises(StopIteration, self.obj_full.read)
        self.obj_sparse.seek(0)
        self.obj_full.seek(0)
        self.test_read()

    def test_write(self):
        for obj in [self.obj_sparse, self.obj_full]:
            w = obj.read()
            f = tempfile.NamedTemporaryFile(suffix=".txt")
            fname = f.name
            f.close()
            o = psopen(fname, "w", "stata_text")
            if obj == self.obj_sparse:
                o.write(w)
            else:
                o.write(w, matrix_form=True)
            o.close()
            wnew = psopen(fname, "r", "stata_text").read()
            assert wnew.pct_nonzero == w.pct_nonzero
            os.remove(fname)



import os
import tempfile

from .... import examples as pysal_examples
from ..pyDbfIO import DBF


class Testtest_DBF:
    def setup_method(self):
        self.test_file = test_file = pysal_examples.get_path("10740.dbf")
        self.dbObj = DBF(test_file, "r")

    def test_len(self):
        assert len(self.dbObj) == 195

    def test_tell(self):
        assert self.dbObj.tell() == 0
        self.dbObj.read(1)
        assert self.dbObj.tell() == 1
        self.dbObj.read(50)
        assert self.dbObj.tell() == 51
        self.dbObj.read()
        assert self.dbObj.tell() == 195

    def test_cast(self):
        assert self.dbObj._spec == []
        self.dbObj.cast("FIPSSTCO", float)
        assert self.dbObj._spec[1] == float

    def test_seek(self):
        self.dbObj.seek(0)
        assert self.dbObj.tell() == 0
        self.dbObj.seek(55)
        assert self.dbObj.tell() == 55
        self.dbObj.read(1)
        assert self.dbObj.tell() == 56

    def test_read(self):
        self.dbObj.seek(0)
        objs = self.dbObj.read()
        assert len(objs) == 195
        self.dbObj.seek(0)
        objsB = list(self.dbObj)
        assert len(objsB) == 195
        for rowA, rowB in zip(objs, objsB):
            assert rowA == rowB

    def test_random_access(self):
        self.dbObj.seek(0)
        db0 = self.dbObj.read(1)[0]
        assert db0 == [1, "35001", "000107", "35001000107", "1.07"]
        self.dbObj.seek(57)
        db57 = self.dbObj.read(1)[0]
        assert db57 == [58, "35001", "001900", "35001001900", "19"]
        self.dbObj.seek(32)
        db32 = self.dbObj.read(1)[0]
        assert db32 == [33, "35001", "000500", "35001000500", "5"]
        self.dbObj.seek(0)
        assert next(self.dbObj) == db0
        self.dbObj.seek(57)
        assert next(self.dbObj) == db57
        self.dbObj.seek(32)
        assert next(self.dbObj) == db32

    def test_write(self):
        f = tempfile.NamedTemporaryFile(suffix=".dbf")
        fname = f.name
        f.close()
        self.dbfcopy = fname
        self.out = DBF(fname, "w")
        self.dbObj.seek(0)
        self.out.header = self.dbObj.header
        self.out.field_spec = self.dbObj.field_spec
        for row in self.dbObj:
            self.out.write(row)
        self.out.close()

        orig = open(self.test_file, "rb")
        copy = open(self.dbfcopy, "rb")
        orig.seek(32)  # self.dbObj.header_size) #skip the header, file date has changed
        copy.seek(32)  # self.dbObj.header_size) #skip the header, file date has changed

        # PySAL writes proper DBF files with a terminator at the end, not everyone does.
        n = self.dbObj.record_size * self.dbObj.n_records  # bytes to read.
        assert orig.read(n) == copy.read(n)
        # self.assertEquals(orig.read(1), copy.read(1)) # last byte may fail
        orig.close()
        copy.close()
        os.remove(self.dbfcopy)

    def test_write_nones(self):
        import datetime
        import time

        f = tempfile.NamedTemporaryFile(suffix=".dbf")
        fname = f.name
        f.close()
        db = DBF(fname, "w")
        db.header = ["recID", "date", "strID", "aFloat"]
        db.field_spec = [("N", 10, 0), ("D", 8, 0), ("C", 10, 0), ("N", 5, 5)]
        records = []
        for i in range(10):
            d = datetime.date(*time.localtime()[:3])
            rec = [i + 1, d, str(i + 1), (i + 1) / 2.0]
            records.append(rec)
        records.append([None, None, "", None])
        records.append(rec)
        for rec in records:
            db.write(rec)
        db.close()
        db2 = DBF(fname, "r")
        assert records == db2.read()
        db2.close()
        os.remove(fname)



"""
sphere: Tools for working with spherical geometry.

Author(s):
    Charles R Schmidt schmidtc@gmail.com
    Luc Anselin luc.anselin@asu.edu
    Xun Li xun.li@asu.edu

"""

__author__ = (
    "Charles R Schmidt <schmidtc@gmail.com>,"
    "Luc Anselin <luc.anselin@asu.edu,"
    "Xun Li <xun.li@asu.edu"
)

import math
from math import cos, pi, sin

import numpy
import scipy.constants
import scipy.spatial
from scipy.spatial.distance import euclidean

__all__ = [
    "RADIUS_EARTH_KM",
    "RADIUS_EARTH_MILES",
    "arcdist",
    "arcdist2linear",
    "brute_knn",
    "fast_knn",
    "fast_threshold",
    "linear2arcdist",
    "toLngLat",
    "toXYZ",
    "lonlat",
    "harcdist",
    "geointerpolate",
    "geogrid",
]


RADIUS_EARTH_KM = 6371.0
RADIUS_EARTH_MILES = (RADIUS_EARTH_KM * scipy.constants.kilo) / scipy.constants.mile


def arcdist(pt0, pt1, radius=RADIUS_EARTH_KM):
    """Arc distance between two points on a sphere.

    Parameters
    ----------
    pt0 : tuple
        A point assumed to be in form (longitude,latitude).
    pt1 : tuple
        A point assumed to be in form (longitude,latitude).
    radius : float
        The radius of a sphere. Default is Earth's radius in
        kilometers, ``RADIUS_EARTH_KM`` (``6371.0``). Earth's
        radius in miles, ``RADIUS_EARTH_MILES`` (``3958.76``)
        is also an option.
        Source: http://nssdc.gsfc.nasa.gov/planetary/factsheet/earthfact.html

    Returns
    -------
    dist : float
        The arc distance between ``pt0`` and ``pt1`` using supplied ``radius``.

    Examples
    --------
    >>> pt0 = (0, 0)
    >>> pt1 = (180, 0)
    >>> d = arcdist(pt0, pt1, RADIUS_EARTH_MILES)
    >>> d == math.pi * RADIUS_EARTH_MILES
    True
    """

    dist = linear2arcdist(euclidean(toXYZ(pt0), toXYZ(pt1)), radius)

    return dist


def arcdist2linear(arc_dist, radius=RADIUS_EARTH_KM):
    """Convert an arc distance (spherical earth)
    to a linear distance (R3) in the unit sphere.

    Parameters
    ----------
    arc_dist : float
        The arc distance to convert.
    radius : float
        The radius of a sphere. Default is Earth's radius in
        kilometers, ``RADIUS_EARTH_KM`` (``6371.0``). Earth's
        radius in miles, ``RADIUS_EARTH_MILES`` (``3958.76``)
        is also an option.
        Source: http://nssdc.gsfc.nasa.gov/planetary/factsheet/earthfact.html

    Returns
    -------
    linear_dist : float
        The linear distance conversion of ``arc_dist``.

    Examples
    --------
    >>> pt0 = (0, 0)
    >>> pt1 = (180, 0)
    >>> d = arcdist(pt0, pt1, RADIUS_EARTH_MILES)
    >>> d == math.pi * RADIUS_EARTH_MILES
    True

    >>> arcdist2linear(d, RADIUS_EARTH_MILES)
    2.0
    """

    circumference = 2 * math.pi * radius
    linear_dist = (
        2 - (2 * math.cos(math.radians((arc_dist * 360.0) / circumference)))
    ) ** (0.5)

    return linear_dist


def linear2arcdist(linear_dist, radius=RADIUS_EARTH_KM):
    """Convert a linear distance in the unit sphere
    (R3) to an arc distance based on supplied radius.

    Parameters
    ----------
    linear_dist : float
        The linear distance to convert.
    radius : float
        The radius of a sphere. Default is Earth's radius in
        kilometers, ``RADIUS_EARTH_KM`` (``6371.0``). Earth's
        radius in miles, ``RADIUS_EARTH_MILES`` (``3958.76``)
        is also an option.
        Source: http://nssdc.gsfc.nasa.gov/planetary/factsheet/earthfact.html

    Returns
    -------
    arc_dist : float
        The arc distance conversion of ``linear_dist``.

    Raises
    ------
    ValueError
        Raised when ``linear_dist`` exceeds the diameter of the unit sphere.

    Examples
    --------
    >>> pt0 = (0, 0)
    >>> pt1 = (180, 0)
    >>> d = arcdist(pt0, pt1, RADIUS_EARTH_MILES)
    >>> d == linear2arcdist(2.0, radius=RADIUS_EARTH_MILES)
    True
    """

    if linear_dist == float("inf"):
        arc_dist = linear_dist
    elif linear_dist > 2.0:
        msg = "'linear_dist', must not exceed the diameter of the unit sphere, 2.0."
        raise ValueError(msg)
    else:
        circumference = 2 * math.pi * radius
        a2 = linear_dist**2
        theta = math.degrees(math.acos((2 - a2) / (2.0)))
        arc_dist = (theta * circumference) / 360.0

    return arc_dist


def toXYZ(pt):  # noqa N802
    """Convert a point's latitude and longitude to x,y,z.

    Parameters
    ----------
    pt : tuple
        A point assumed to be in form (lng,lat).

    Returns
    -------
    x, y, z : tuple
        A point in form (x, y, z).
    """

    phi, theta = list(map(math.radians, pt))
    phi, theta = phi + pi, theta + (pi / 2)
    x = 1 * sin(theta) * cos(phi)
    y = 1 * sin(theta) * sin(phi)
    z = 1 * cos(theta)

    return x, y, z


def toLngLat(xyz):  # noqa N802
    """Convert a point's x,y,z to latitude and longitude.

    Parameters
    ----------
    xyz : tuple
        A point assumed to be in form (x,y,z).

    Returns
    -------
    phi, theta : tuple
        A point in form (phi, theta) [y,x].
    """

    x, y, z = xyz
    if z == -1 or z == 1:
        phi = 0
    else:
        phi = math.atan2(y, x)
        if phi > 0:
            phi = phi - math.pi
        elif phi < 0:
            phi = phi + math.pi
    theta = math.acos(z) - (math.pi / 2)

    return phi, theta


def brute_knn(pts, k, mode="arc", radius=RADIUS_EARTH_KM):
    """Computes a brute-force :math:`k` nearest neighbors.

    Parameters
    ----------
    pts : list
        A list of :math:`x,y` pairs.
    k :  int
        The number of points to query.
    mode : str
        The mode of distance. Valid modes are ``'arc'``
        and ``'xyz'``. Default is ``'arc'``.
    radius : float
        The radius of a sphere. Default is Earth's radius in
        kilometers, ``RADIUS_EARTH_KM`` (``6371.0``). Earth's
        radius in miles, ``RADIUS_EARTH_MILES`` (``3958.76``)
        is also an option.
        Source: http://nssdc.gsfc.nasa.gov/planetary/factsheet/earthfact.html

    Returns
    -------
    w : dict
        A neighbor ID lookup.
    """

    n = len(pts)
    full = numpy.zeros((n, n))

    for i in range(n):
        for j in range(i + 1, n):
            if mode == "arc":
                lng0, lat0 = pts[i]
                lng1, lat1 = pts[j]
                dist = arcdist(pts[i], pts[j], radius=radius)
            elif mode == "xyz":
                dist = euclidean(pts[i], pts[j])
            full[i, j] = dist
            full[j, i] = dist

    w = {}
    for i in range(n):
        w[i] = full[i].argsort()[1 : k + 1].tolist()

    return w


def fast_knn(pts, k, return_dist=False, radius=RADIUS_EARTH_KM):
    """Computes :math:`k` nearest neighbors on a sphere.

    Parameters
    ----------
    pts : list
        A list of :math:`x,y` pairs.
    k :  int
        The number of points to query.
    return_dist : bool
        Return distances in the ``wd`` container object (``True``).
        Default is ``False``.
    radius : float
        The radius of a sphere. Default is Earth's radius in
        kilometers, ``RADIUS_EARTH_KM`` (``6371.0``). Earth's
        radius in miles, ``RADIUS_EARTH_MILES`` (``3958.76``)
        is also an option.
        Source: http://nssdc.gsfc.nasa.gov/planetary/factsheet/earthfact.html

    Returns
    -------
    wn : dict
        A neighbor ID lookup.
    wd : dict
        A neighbor distance lookup (optional).
    """

    pts = numpy.array(pts)
    kd = scipy.spatial.KDTree(pts)
    d, w = kd.query(pts, k + 1)
    w = w[:, 1:]
    wn = {}

    for i in range(len(pts)):
        wn[i] = w[i].tolist()

    if return_dist:
        d = d[:, 1:]
        wd = {}
        for i in range(len(pts)):
            wd[i] = [linear2arcdist(x, radius=radius) for x in d[i].tolist()]
        return wn, wd
    return wn


def fast_threshold(pts, dist, radius=RADIUS_EARTH_KM):
    """Find all neighbors on a sphere within a threshold distance.

    Parameters
    ----------
    pointslist : list
        A list of lat-lon tuples. This **must** be a list, even for one point.
    dist: float
        The threshold distance.
    radius : float
        The radius of a sphere. Default is Earth's radius in
        kilometers, ``RADIUS_EARTH_KM`` (``6371.0``). Earth's
        radius in miles, ``RADIUS_EARTH_MILES`` (``3958.76``)
        is also an option.
        Source: http://nssdc.gsfc.nasa.gov/planetary/factsheet/earthfact.html

    Returns
    -------
    wd : dict
        A neighbor distance lookup where the key is the ID
        of a point and the value is a list of IDs for other
        points within ``dist`` of the key point,
    """

    d = arcdist2linear(dist, radius)
    kd = scipy.spatial.KDTree(pts)
    r = kd.query_ball_tree(kd, d)
    wd = {}

    for i in range(len(pts)):
        l_ = r[i]
        l_.remove(i)
        wd[i] = l_

    return wd


def lonlat(pointslist):
    """Converts point order from lat-lon tuples to lon-lat (x,y) tuples.

    Parameters
    ----------
    pointslist : list
        A list of lat-lon tuples. This **must** be a list, even for one point.

    Returns
    -------
    newpts : list
        A list with tuples of points in lon-lat order.

    Examples
    --------
    >>> points = [
    ...     (41.981417, -87.893517), (41.980396, -87.776787), (41.980906, -87.696450)
    ... ]
    >>> newpoints = lonlat(points)
    >>> newpoints
    [(-87.893517, 41.981417), (-87.776787, 41.980396), (-87.69645, 41.980906)]
    """

    newpts = [(i[1], i[0]) for i in pointslist]

    return newpts


def haversine(x):
    """Computes the haversine formula.

    Parameters
    ----------
    x : float
        The angle in radians.

    Returns
    -------
    haversine_dist : float
        The square of sine of half the radian (the haversine formula).

    Examples
    --------
    >>> haversine(math.pi) # is 180 in radians, hence sin of 90 = 1
    1.0
    """

    x = math.sin(x / 2)

    haversine_dist = x * x

    return haversine_dist


# Lambda functions

# degree to radian conversion
d2r = lambda x: x * math.pi / 180.0

# radian to degree conversion
r2d = lambda x: x * 180.0 / math.pi


def radangle(p0, p1):
    """Radian angle between two points on a sphere in lon-lat (x,y).

    Parameters
    ----------
    p0 : tuple
        The first point in (lon,lat) format.
    p1 : tuple
        The second point in (lon,lat) format.

    Returns
    -------
    d : float
        Radian angle in radians.

    Examples
    --------
    >>> p0 = (-87.893517, 41.981417)
    >>> p1 = (-87.519295, 41.657498)
    >>> radangle(p0, p1)
    0.007460167953189258

    Notes
    -----
    Uses haversine formula, function haversine and degree to radian
    conversion lambda function ``d2r``.
    """

    x0, y0 = d2r(p0[0]), d2r(p0[1])
    x1, y1 = d2r(p1[0]), d2r(p1[1])
    d = 2.0 * math.asin(
        math.sqrt(haversine(y1 - y0) + math.cos(y0) * math.cos(y1) * haversine(x1 - x0))
    )

    return d


def harcdist(p0, p1, lonx=True, radius=RADIUS_EARTH_KM):
    """Alternative the arc distance function, uses the haversine formula.

    Parameters
    ----------
    p0 : tuple
        The first point decimal degrees.
    p1 : tuple
        The second point decimal degrees.
    lonx : bool
        The method to assess the order of the coordinates.
        ``True`` for (lon,lat); ``False`` for (lat,lon).
        Default is ``True``.
    radius : float
        The radius of a sphere. Default is Earth's radius in
        kilometers, ``RADIUS_EARTH_KM`` (``6371.0``). Earth's
        radius in miles, ``RADIUS_EARTH_MILES`` (``3958.76``)
        is also an option. Set to ``None`` for radians.
        Source: http://nssdc.gsfc.nasa.gov/planetary/factsheet/earthfact.html

    Returns
    -------
    harc_dist : harc_dist
        The distance in units specified, km, miles or radians.

    Examples
    --------
    >>> p0 = (-87.893517, 41.981417)
    >>> p1 = (-87.519295, 41.657498)
    >>> harcdist(p0, p1)
    47.52873002976876

    >>> harcdist(p0, p1, radius=None)
    0.007460167953189258

    Notes
    -----
    Uses the ``radangle`` function to compute radian angle.
    """

    if not (lonx):
        p = lonlat([p0, p1])
        p0 = p[0]
        p1 = p[1]

    harc_dist = radangle(p0, p1)

    if radius is not None:
        harc_dist = harc_dist * radius

    return harc_dist


def geointerpolate(p0, p1, t, lonx=True):
    r"""Finds a point on a sphere along the great circle distance between
    two points on a sphere also known as a way point in great circle navigation.

    Parameters
    ----------
    p0 : tuple
        The first point decimal degrees.
    p1 : tuple
        The second point decimal degrees.
    t : float
        The proportion along great circle distance between ``p0``
        and ``p1`` (e.g., :math:`\mathtt{t}=0.5` would find the mid-point).
    lonx : bool
        The method to assess the order of the coordinates.
        ``True`` for (lon,lat); ``False`` for (lat,lon).
        Default is ``True``.

    Returns
    -------
    newpx, newpy : tuple
        The new point in decimal degrees of (lon-lat) by
        default or (lat-lon) if ``lonx`` is set to ``False``.

    Examples
    --------
    >>> p0 = (-87.893517, 41.981417)
    >>> p1 = (-87.519295, 41.657498)
    >>> geointerpolate(p0, p1, 0.1)             # using lon-lat
    (-87.85592403438788, 41.949079912574796)

    >>> p3 = (41.981417, -87.893517)
    >>> p4 = (41.657498, -87.519295)
    >>> geointerpolate(p3, p4, 0.1, lonx=False) # using lat-lon
    (41.949079912574796, -87.85592403438788)
    """

    if not (lonx):
        p = lonlat([p0, p1])
        p0 = p[0]
        p1 = p[1]

    d = radangle(p0, p1)
    k = 1.0 / math.sin(d)
    t = t * d
    A = math.sin(d - t) * k
    B = math.sin(t) * k

    x0, y0 = d2r(p0[0]), d2r(p0[1])
    x1, y1 = d2r(p1[0]), d2r(p1[1])

    x = A * math.cos(y0) * math.cos(x0) + B * math.cos(y1) * math.cos(x1)
    y = A * math.cos(y0) * math.sin(x0) + B * math.cos(y1) * math.sin(x1)
    z = A * math.sin(y0) + B * math.sin(y1)

    newpx = r2d(math.atan2(y, x))
    newpy = r2d(math.atan2(z, math.sqrt(x * x + y * y)))

    if not lonx:
        return newpy, newpx

    return newpx, newpy


def geogrid(pup, pdown, k, lonx=True):
    """Computes a :math:`k+1` by :math:`k+1` set of grid
    points for a bounding box in lat-lon. Uses ``geointerpolate``.

    Parameters
    ----------
    pup : tuple
        The lat-lon or lon-lat for the upper left corner of the bounding box.
    pdown : tuple
        The lat-lon or lon-lat for The lower right corner of The bounding box.
    k : int
        The number of grid cells (grid points will be one more).
    lonx : bool
        The method to assess the order of the coordinates.
        ``True`` for (lon,lat); ``False`` for (lat,lon).
        Default is ``True``.

    Returns
    -------
    grid : list
        A list of tuples with (lat-lon) or (lon-lat) for grid points,
        row by row, starting with the top row and moving to the bottom;
        coordinate tuples are returned in same order as input.

    Examples
    --------
    >>> pup = (42.023768, -87.946389)       # Arlington Heights, IL
    >>> pdown = (41.644415, -87.524102)     # Hammond, IN
    >>> geogrid(pup,pdown, 3, lonx=False)
    [(42.023768, -87.946389),
     (42.02393997819538, -87.80562679358316),
     (42.02393997819538, -87.66486420641684),
     (42.023768, -87.524102),
     (41.897317, -87.94638900000001),
     (41.8974888973743, -87.80562679296166),
     (41.8974888973743, -87.66486420703835),
     (41.897317, -87.524102),
     (41.770866000000005, -87.94638900000001),
     (41.77103781320412, -87.80562679234043),
     (41.77103781320412, -87.66486420765956),
     (41.770866000000005, -87.524102),
     (41.644415, -87.946389),
     (41.64458672568646, -87.80562679171955),
     (41.64458672568646, -87.66486420828045),
     (41.644415, -87.524102)]
    """

    corners = [pup, pdown] if lonx else lonlat([pup, pdown])

    tpoints = [float(i) / k for i in range(k)[1:]]
    leftcorners = [corners[0], (corners[0][0], corners[1][1])]
    rightcorners = [(corners[1][0], corners[0][1]), corners[1]]
    leftside = [leftcorners[0]]
    rightside = [rightcorners[0]]

    for t in tpoints:
        newpl = geointerpolate(leftcorners[0], leftcorners[1], t)
        leftside.append(newpl)
        newpr = geointerpolate(rightcorners[0], rightcorners[1], t)
        rightside.append(newpr)
    leftside.append(leftcorners[1])
    rightside.append(rightcorners[1])

    grid = []
    for i in range(len(leftside)):
        grid.append(leftside[i])
        for t in tpoints:
            newp = geointerpolate(leftside[i], rightside[i], t)
            grid.append(newp)
        grid.append(rightside[i])
    if not (lonx):
        grid = lonlat(grid)

    return grid



from ... import examples as pysal_examples
from ...io.fileio import FileIO as psopen  # noqa N813
from ..shapes import Chain, Point, asShape


class TesttestMultiPloygon:
    def test___init__1(self):
        """Tests conversion of polygons with multiple shells to
        geoJSON multipolygons and back.

        """

        shp = psopen(pysal_examples.get_path("NAT.shp"), "r")
        multipolygons = [p for p in shp if len(p.parts) > 1]
        for poly in multipolygons:
            json = poly.__geo_interface__
            shape = asShape(json)
            assert json["type"] == "MultiPolygon"
            assert str(shape.holes) == str(poly.holes)
            assert str(shape.parts) == str(poly.parts)


class TesttestMultiLineString:
    def test_multipart_chain(self):
        vertices = [
            [Point((0, 0)), Point((1, 0)), Point((1, 5))],
            [Point((-5, -5)), Point((-5, 0)), Point((0, 0))],
        ]

        # part A
        chain0 = Chain(vertices[0])
        # part B
        chain1 = Chain(vertices[1])
        # part A and B
        chain2 = Chain(vertices)

        json = chain0.__geo_interface__
        assert json["type"] == "LineString"
        assert len(json["coordinates"]) == 3

        json = chain1.__geo_interface__
        assert json["type"] == "LineString"
        assert len(json["coordinates"]) == 3

        json = chain2.__geo_interface__
        assert json["type"] == "MultiLineString"
        assert len(json["coordinates"]) == 2

        chain3 = asShape(json)
        assert chain2.parts == chain3.parts
